"""Manually specified (i.e. not learned) models."""
import pandas as pd
import numpy as np

from ethicml.common import implements
from ethicml.algorithms.inprocess.in_algorithm import InAlgorithm
from ethicml.utility.data_structures import DataTuple, TestTuple


class Corels(InAlgorithm):
    """CORELS (Certifiably Optimal RulE ListS) algorithm for the COMPAS dataset.

    This algorithm uses if-statements to make predictions. It only works on COMPAS with s as sex.

    From this paper: https://arxiv.org/abs/1704.01701
    """

    @implements(InAlgorithm)
    def run(self, _: DataTuple, test: TestTuple) -> pd.DataFrame:
        if test.name is None or "Compas" not in test.name or "sex" not in test.s.columns:
            raise RuntimeError("The Corels algorithm only works on the COMPAS dataset")
        age = test.x["age-num"].to_numpy()
        priors = test.x["priors-count"].to_numpy()
        sex = test.s["sex"].to_numpy()
        male = 1
        condition1 = (age >= 18) & (age <= 20) & (sex == male)
        condition2 = (age >= 21) & (age <= 23) & (priors >= 2) & (priors <= 3)
        condition3 = priors > 3
        pred = np.where(condition1 | condition2 | condition3, np.ones_like(age), np.zeros_like(age))
        return pd.DataFrame(pred, columns=["preds"])

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return "CORELS"



"""Common plotting functions / datastructures."""

from typing import Tuple, NamedTuple, List, Optional, Dict, Any
from typing_extensions import Literal

import pandas as pd
from matplotlib import pyplot as plt
import matplotlib as mpl

LegendType = Literal["inside", "outside"]  # pylint: disable=invalid-name
PlotType = Literal["box", "cross", "scatter", "line"]  # pylint: disable=invalid-name


class DataEntry(NamedTuple):
    """Defines one element in a plot."""

    label: str
    values: pd.DataFrame
    do_fill: bool


class PlotDef(NamedTuple):
    """All the things needed to make a plot.

    Attributes:
        title: the title of the plot
        entries: a List of DataEntries
        legend_pos: where to put the legend; allowed values: None, "inside", "outside"
        legend_yanchor: float that specifies the vertical position of the legend
    """

    title: str
    entries: List[DataEntry]
    legend_pos: Optional[LegendType] = None
    legend_yanchor: float = 1.0


def common_plotting_settings(
    plot: plt.Axes, plot_def: PlotDef, xaxis_title: str, yaxis_title: str
) -> Optional[mpl.legend.Legend]:
    """Common settings for plots.

    Args:
        plot: a pyplot plot object
        plot_def: a `PlotDef` that defines properties of the plot
        xaxis_title: label for x-axis
        yaxis_title: label for y-axis
    """
    plot.set_xlabel(xaxis_title)
    plot.set_ylabel(yaxis_title)
    if plot_def.title:
        plot.set_title(plot_def.title)
    plot.grid(True)

    # get handles and labels for the legend
    handles, labels = plot.get_legend_handles_labels()
    # remove the errorbars from the legend if they are there
    handles = [(h[0] if isinstance(h, tuple) else h) for h in handles]

    if plot_def.legend_pos == "inside":
        return plot.legend(handles, labels)
    if plot_def.legend_pos == "outside":
        return plot.legend(
            handles, labels, loc='upper left', bbox_to_anchor=(1, plot_def.legend_yanchor)
        )
    return None


def errorbox(
    plot: plt.Axes,
    plot_def: PlotDef,
    xaxis: Tuple[str, str],
    yaxis: Tuple[str, str],
    firstcolor: int = 0,
    firstshape: int = 0,
    markersize: int = 6,
    use_cross: bool = False,
) -> Optional[mpl.legend.Legend]:
    """Generate a figure with errorboxes that reflect the std dev of an entry.

    Args:
        plot: a pyplot Axes object
        plot_def: a `PlotDef` that defines properties of the plot
        xaxis: a tuple of two strings
        yaxis: a tuple of two strings
        firstcolor: index of the color that should be used for the first entry
        firstshape: index of the shape that should be used for the first entry
        markersize: size of the markers
        use_cross: if True, use cross instead of boxes
    """
    # ================================ constants for plots ========================================
    colors = ["#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd"]
    colors += ["#8c564b", "#e377c2", "#7f7f7f", "#bcbd22", "#17becf"]
    pale_colors = ["#aec7e8", "#ffbb78", "#98df8a", "#ff9896", "#c5b0d5"]
    pale_colors += ["#c49c94", "#f7b6d2", "#c7c7c7", "#dbdb8d", "#9edae5"]
    shapes = ['o', 'X', 'D', 's', '^', 'v', '<', '>', '*', 'p', 'P']
    hatch_pattern = ['O', 'o', '|', '-', '/', '\\', '+', 'x', '*']

    xaxis_measure, yaxis_measure = xaxis[0], yaxis[0]
    filled_counter = firstcolor

    max_xmean = 0.0
    min_xmean = 1.0

    max_ymean = 0.0
    min_ymean = 1.0

    for i, entry in enumerate(plot_def.entries):
        if entry.do_fill:
            color = colors[filled_counter]
            filled_counter += 1
        else:
            color = pale_colors[i + firstcolor - filled_counter]
        i_shp = firstshape + i

        xmean: float = entry.values[xaxis_measure].mean()
        xstd: float = entry.values[xaxis_measure].std()
        if xmean + (0.5 * xstd) > max_xmean:
            max_xmean = xmean + (0.5 * xstd)
        if xmean - (0.5 * xstd) < min_xmean:
            min_xmean = xmean - (0.5 * xstd)

        ymean: float = entry.values[yaxis_measure].mean()
        ystd: float = entry.values[yaxis_measure].std()
        if ymean + (0.5 * ystd) > max_ymean:
            max_ymean = ymean + (0.5 * ystd)
        if ymean - (0.5 * ystd) < min_ymean:
            min_ymean = ymean - (0.5 * ystd)

        if use_cross:
            plot.errorbar(
                xmean,
                ymean,
                xerr=0.5 * xstd,
                yerr=0.5 * ystd,
                marker=shapes[i_shp % len(shapes)],
                linestyle="",  # no connecting lines
                color=color,
                ecolor="#555555",
                capsize=4,
                elinewidth=1.0,
                zorder=3 + 2 * i_shp,
                label=entry.label,
                markersize=markersize,
            )
        else:
            plot.bar(
                xmean,
                ystd,
                bottom=ymean - 0.5 * ystd,
                width=xstd,
                align='center',
                color='none',
                edgecolor=color,
                linewidth=3,
                zorder=3 + 2 * i_shp,
                hatch=hatch_pattern[i % len(hatch_pattern)],
            )
            plot.plot(
                xmean,
                ymean,
                marker=shapes[i_shp % len(shapes)],
                linestyle="",  # no connecting lines
                color=color,
                label=entry.label,
                zorder=4 + 2 * i_shp,
                markersize=markersize,
            )

    x_min = min_xmean * 0.99
    x_max = max_xmean * 1.01

    y_min = min_ymean * 0.99
    y_max = max_ymean * 1.01

    plot.set_xlim(x_min, x_max)
    plot.set_ylim(y_min, y_max)
    return common_plotting_settings(plot, plot_def, xaxis[1], yaxis[1])


def scatter(
    plot: plt.Axes,
    plot_def: PlotDef,
    xaxis: Tuple[str, str],
    yaxis: Tuple[str, str],
    startindex: int = 0,
    markersize: int = 6,
    connect_dots: bool = False,
) -> Optional[mpl.legend.Legend]:
    """Generate a scatter plot.

    Args:
        plot: a pyplot plot object
        plot_def: a `PlotDef` that defines properties of the plot
        xaxis: a tuple of two strings
        yaxis: a tuple of two strings
        startindex: index of color and shape that should be used for the first entry
        markersize: size of the markers
        connect_dots: if True, connect the data points with a line
    """
    shapes = ['o', 'X', 'D', 's', '^', 'v', '<', '>', '*', 'p', 'P']
    colors10 = ["#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd"]
    colors10 += ["#8c564b", "#e377c2", "#7f7f7f", "#bcbd22", "#17becf"]
    xaxis_measure, yaxis_measure = xaxis[0], yaxis[0]
    filled_counter = startindex
    for i, entry in enumerate(plot_def.entries):
        additional_params: Dict[str, Any]
        if entry.do_fill or connect_dots:
            additional_params = dict()
            shp_index = filled_counter
            filled_counter += 1
        else:
            additional_params = dict(markerfacecolor='none')
            shp_index = i + startindex - filled_counter
        plot.plot(
            entry.values[xaxis_measure].to_numpy(),
            entry.values[yaxis_measure].to_numpy(),
            marker=shapes[shp_index],
            linestyle="-" if connect_dots else "",
            label=entry.label,
            color=None if connect_dots else colors10[shp_index],
            markersize=markersize,
            **additional_params,
        )
    return common_plotting_settings(plot, plot_def, xaxis[1], yaxis[1])



"""Data structures that are used throughout the code."""
from pathlib import Path
from dataclasses import dataclass
from typing import Tuple, List, Optional, NamedTuple, Callable, Iterator
from typing_extensions import Literal, Final

import pandas as pd
from pandas.testing import assert_index_equal

AxisType = Literal["columns", "index"]  # pylint: disable=invalid-name


class TestTuple:
    """A tuple of dataframes for the features and the sensitive attribute."""

    def __init__(self, x: pd.DataFrame, s: pd.DataFrame, name: Optional[str] = None):
        """Init function for TestTuple object."""
        self.__x: pd.DataFrame = x
        self.__s: pd.DataFrame = s
        self.__name: Optional[str] = name

    @property
    def x(self) -> pd.DataFrame:
        """Getter for property x."""
        return self.__x

    @property
    def s(self) -> pd.DataFrame:
        """Getter for property s."""
        return self.__s

    @property
    def name(self) -> Optional[str]:
        """Getter for name property."""
        return self.__name

    def __iter__(self) -> Iterator[pd.DataFrame]:
        """Overwrite magic method __iter__."""
        return iter([self.x, self.s])

    def write_as_feather(self, data_dir: Path, identifier: str) -> "TestPathTuple":
        """Write the TestTuple to Feather files and return the file paths as a TestPathTuple.

        Args:
            data_dir: directory where the files should be stored
            identifier: a string that ideally identifies the file uniquely, can be a UUID
        Returns:
            tuple of paths
        """
        # create the directory if it doesn't already exist
        data_dir.mkdir(parents=True, exist_ok=True)

        return TestPathTuple(
            x=_save_helper(data_dir, self.x, identifier, "x"),
            s=_save_helper(data_dir, self.s, identifier, "s"),
            name=self.name if self.name is not None else "",
        )

    def replace(
        self,
        *,
        x: Optional[pd.DataFrame] = None,
        s: Optional[pd.DataFrame] = None,
        name: Optional[str] = None,
    ) -> "TestTuple":
        """Create a copy of the TestTuple but change the given values."""
        return TestTuple(
            x=x if x is not None else self.x,
            s=s if s is not None else self.s,
            name=name if name is not None else self.name,
        )


class DataTuple(TestTuple):
    """A tuple of dataframes for the features, the sensitive attribute and the class labels."""

    def __init__(
        self, x: pd.DataFrame, s: pd.DataFrame, y: pd.DataFrame, name: Optional[str] = None
    ):
        """Init for DataTuple Class."""
        super().__init__(x=x, s=s, name=name)
        self.__y: pd.DataFrame = y

    @property
    def y(self) -> pd.DataFrame:
        """Getter for property y."""
        return self.__y

    def __iter__(self) -> Iterator[pd.DataFrame]:
        """Overwrite __iter__ magic method."""
        return iter([self.x, self.s, self.y])

    def __len__(self) -> int:
        """Overwrite __len__ magic method."""
        len_x = len(self.x)
        assert len_x == len(self.s) and len_x == len(self.y)
        return len_x

    def remove_y(self) -> TestTuple:
        """Convert the DataTuple instance to a TestTuple instance."""
        return TestTuple(x=self.x, s=self.s, name=self.name)

    def write_as_feather(self, data_dir: Path, identifier: str) -> "PathTuple":
        """Write the DataTuple to Feather files and return the file paths as a PathTuple.

        Args:
            data_dir: directory where the files should be stored
            identifier: a string that ideally identifies the file uniquely, can be a UUID
        Returns:
            tuple of paths
        """
        # create the directory if it doesn't already exist
        data_dir.mkdir(parents=True, exist_ok=True)

        return PathTuple(
            x=_save_helper(data_dir, self.x, identifier, "x"),
            s=_save_helper(data_dir, self.s, identifier, "s"),
            y=_save_helper(data_dir, self.y, identifier, "y"),
            name=self.name if self.name is not None else "",
        )

    def replace(
        self,
        *,
        x: Optional[pd.DataFrame] = None,
        s: Optional[pd.DataFrame] = None,
        y: Optional[pd.DataFrame] = None,
        name: Optional[str] = None,
    ) -> "DataTuple":
        """Create a copy of the DataTuple but change the given values."""
        return DataTuple(
            x=x if x is not None else self.x,
            s=s if s is not None else self.s,
            y=y if y is not None else self.y,
            name=name if name is not None else self.name,
        )

    def apply_to_joined_df(self, mapper: Callable[[pd.DataFrame], pd.DataFrame]) -> "DataTuple":
        """Concatenate the dataframes in the DataTuple and then apply a function to it."""
        self.x.columns = self.x.columns.astype(str)
        cols_x, cols_s, cols_y = self.x.columns, self.s.columns, self.y.columns
        joined = pd.concat([self.x, self.s, self.y], axis="columns", sort=False)
        assert len(joined) == len(self), "something went wrong while concatenating"
        joined = mapper(joined)
        result = self.replace(x=joined[cols_x], s=joined[cols_s], y=joined[cols_y])

        # assert that the columns haven't changed
        assert_index_equal(result.x.columns, cols_x)
        assert_index_equal(result.s.columns, cols_s)
        assert_index_equal(result.y.columns, cols_y)

        return result

    def get_subset(self, num: int = 500) -> "DataTuple":
        """Get the first elements of the dataset.

        Args:
            num: how many samples to take for subset

        Returns:
            subset of training data
        """
        return self.replace(x=self.x.iloc[:num], s=self.s.iloc[:num], y=self.y.iloc[:num])


@dataclass(frozen=True)  # "frozen" means the objects are immutable
class TestPathTuple:
    """For algorithms that run in their own process, we pass around paths to the data."""

    x: Path  # path to file with features
    s: Path  # path to file with sensitive attributes
    name: str  # name of the dataset

    def load_from_feather(self) -> TestTuple:
        """Load a dataframe from a feather file."""
        return TestTuple(
            x=load_feather(self.x), s=load_feather(self.s), name=self.name if self.name else None
        )


@dataclass(frozen=True)  # "frozen" means the objects are immutable
class PathTuple(TestPathTuple):
    """For algorithms that run in their own process, we pass around paths to the data."""

    y: Path  # path to file with class labels

    def load_from_feather(self) -> DataTuple:
        """Load a dataframe from a feather file."""
        return DataTuple(
            x=load_feather(self.x),
            s=load_feather(self.s),
            y=load_feather(self.y),
            name=self.name if self.name else None,
        )


def _save_helper(data_dir: Path, data: pd.DataFrame, prefix: str, key: str) -> Path:
    # SUGGESTION: maybe the file names should be completely random to avoid collisions
    data_path = data_dir / f"data_{prefix}_{key}.feather"
    # write the file
    data.to_feather(data_path)
    return data_path


def write_as_feather(
    train: DataTuple, test: TestTuple, data_dir: Path
) -> (Tuple[PathTuple, TestPathTuple]):
    """Write the given DataTuple to Feather files and return the file paths as PathTuples.

    Args:
        train: tuple with training data
        test: tuple with test data
        data_dir: directory where the files should be stored
    Returns:
        tuple of tuple of paths (one tuple for training, one for test)
    """
    # TODO: this should be an assert instead (requires changing code that calls this function)
    if isinstance(test, DataTuple):
        # because of polymorphism it can happen that `test` is a DataTuple posing as a TestTuple
        # this causes problems though because it will write an additional file (the one with y)
        test = test.remove_y()
    return train.write_as_feather(data_dir, "train"), test.write_as_feather(data_dir, "test")


def concat_dt(
    datatup_list: List[DataTuple], axis: AxisType = "index", ignore_index: bool = False
) -> DataTuple:
    """Concatenate the data tuples in the given list."""
    return DataTuple(
        x=pd.concat(
            [dt.x for dt in datatup_list], axis=axis, sort=False, ignore_index=ignore_index
        ),
        s=pd.concat(
            [dt.s for dt in datatup_list], axis=axis, sort=False, ignore_index=ignore_index
        ),
        y=pd.concat(
            [dt.y for dt in datatup_list], axis=axis, sort=False, ignore_index=ignore_index
        ),
        name=datatup_list[0].name,
    )


def concat_tt(
    datatup_list: List[TestTuple], axis: AxisType = "index", ignore_index: bool = False
) -> TestTuple:
    """Concatenate the test tuples in the given list."""
    return TestTuple(
        x=pd.concat(
            [dt.x for dt in datatup_list], axis=axis, sort=False, ignore_index=ignore_index
        ),
        s=pd.concat(
            [dt.s for dt in datatup_list], axis=axis, sort=False, ignore_index=ignore_index
        ),
        name=datatup_list[0].name,
    )


def load_feather(output_path: Path) -> pd.DataFrame:
    """Load a dataframe from a feather file."""
    with output_path.open("rb") as file_obj:
        df = pd.read_feather(file_obj)
    return df


FairnessType = Literal["DP", "EqOp", "EqOd"]  # pylint: disable=invalid-name


def str_to_fair_type(fair_str: str) -> Optional[FairnessType]:
    """Convert a string to a fairness type or return None if not possible."""
    # this somewhat silly code is needed because mypy doesn't support narrowing to literals yet
    if fair_str == "DP":
        return "DP"
    if fair_str == "EqOd":
        return "EqOd"
    if fair_str == "EqOp":
        return "EqOp"
    return None


ClassifierType = Literal["LR", "SVM"]  # pylint: disable=invalid-name
ActivationType = Literal["identity", "logistic", "tanh", "relu"]  # pylint: disable=invalid-name


class TrainTestPair(NamedTuple):
    """2-Tuple of train and test data."""

    train: DataTuple
    test: TestTuple


class Results:
    """Container for results from `evaluate_models`."""

    columns: Final = ["dataset", "transform", "model", "split_id"]

    def __init__(self, data_frame: Optional[pd.DataFrame] = None):
        """Initialise Results object."""
        self._data: pd.DataFrame
        if data_frame is not None:
            # ensure correct index
            if data_frame.index.names != self.columns:
                self._data = data_frame.set_index(self.columns)
            else:
                self._data = data_frame
        else:
            self._data = pd.DataFrame(columns=self.columns).set_index(self.columns)

    @property
    def data(self) -> pd.DataFrame:
        """Getter for data property."""
        return self._data

    def __repr__(self) -> str:
        """Overwrite __repr__ magic method."""
        return repr(self._data)

    def __str__(self) -> str:
        """Overwrite __str__ magic method."""
        return str(self._data)

    def append_df(self, data_frame: pd.DataFrame, prepend: bool = False) -> None:
        """Append (or prepend) a DataFrame to this object."""
        if data_frame.index.names != self.columns:
            data_frame = data_frame.set_index(self.columns)  # set correct index
        order = [data_frame, self._data] if prepend else [self._data, data_frame]
        # set sort=False so that the order of the columns is preserved
        self._data = pd.concat(order, sort=False, axis="index")

    def append_from_file(self, csv_file: Path, prepend: bool = False) -> bool:
        """Append results from a CSV file."""
        if csv_file.is_file():  # if file exists
            self.append_df(pd.read_csv(csv_file), prepend=prepend)
            return True
        return False

    def save_as_csv(self, file_path: Path) -> None:
        """Save to csv."""
        # `_data` has the multi index based on [dataset, transform, ...] so we have to reset that
        self._data.reset_index(drop=False).to_csv(file_path, index=False)

    @classmethod
    def from_file(cls, csv_file: Path) -> Optional["Results"]:
        """Load results from a CSV file that was created by `evaluate_models`.

        Args:
            csv_file: path to a CSV file with results

        Returns:
            DataFrame if the file exists; None otherwise
        """
        if csv_file.is_file():
            return cls(pd.read_csv(csv_file))
        return None

    def map_over_index(
        self, mapper: Callable[[Tuple[str, str, str, str]], Tuple[str, str, str, str]]
    ) -> pd.DataFrame:
        """Change the values of the index with a transformation function."""
        results_mapped = self._data.copy()
        results_mapped.index = results_mapped.index.map(mapper)
        return results_mapped



"""
Test the loading data capability
"""
from pathlib import Path
import pandas as pd

from ethicml.common import ROOT_PATH
from ethicml.data import (
    Dataset,
    load_data,
    create_data_obj,
    Adult,
    Compas,
    Credit,
    German,
    Sqf,
    Toy,
    NonBinaryToy,
)
from ethicml.utility import DataTuple, concat_dt
from ethicml.preprocessing import domain_split, query_dt


def test_can_load_test_data():
    """test whether we can load test data"""
    data_loc: Path = ROOT_PATH / "data" / "csvs" / "toy.csv"
    data: pd.DataFrame = pd.read_csv(data_loc)
    assert data is not None


def test_load_data():
    """test loading data"""
    data: DataTuple = load_data(Toy())
    assert (2000, 2) == data.x.shape
    assert (2000, 1) == data.s.shape
    assert (2000, 1) == data.y.shape
    assert data.name == "Toy"


def test_load_non_binary_data():
    """test load non binary data"""
    data: DataTuple = load_data(NonBinaryToy())
    assert (100, 2) == data.x.shape
    assert (100, 1) == data.s.shape
    assert (100, 1) == data.y.shape


def test_discrete_data():
    """test discrete data"""
    data: DataTuple = load_data(Adult())
    assert (45222, 101) == data.x.shape
    assert (45222, 1) == data.s.shape
    assert (45222, 1) == data.y.shape
    assert len(Adult().discrete_features) == 96
    assert len(Adult(split='Race').discrete_features) == 93
    assert len(Adult(split='Nationality').discrete_features) == 57


def test_load_data_as_a_function():
    """test load data as a function"""
    data_loc: Path = ROOT_PATH / "data" / "csvs" / "toy.csv"
    data_obj: Dataset = create_data_obj(data_loc, s_columns=["s"], y_columns=["y"])
    assert data_obj is not None
    assert data_obj.feature_split['x'] == ['a1', 'a2']
    assert data_obj.feature_split['s'] == ['s']
    assert data_obj.feature_split['y'] == ['y']
    assert data_obj.filename == "toy.csv"


def test_joining_2_load_functions():
    """test joining 2 load functions"""
    data_loc: Path = ROOT_PATH / "data" / "csvs" / "toy.csv"
    data_obj: Dataset = create_data_obj(data_loc, s_columns=["s"], y_columns=["y"])
    data: DataTuple = load_data(data_obj)
    assert (2000, 2) == data.x.shape
    assert (2000, 1) == data.s.shape
    assert (2000, 1) == data.y.shape


def test_load_adult():
    """test load adult"""
    data: DataTuple = load_data(Adult())
    assert (45222, 101) == data.x.shape
    assert (45222, 1) == data.s.shape
    assert (45222, 1) == data.y.shape
    assert data.name == "Adult Sex"


def test_load_compas():
    """test load compas"""
    data: DataTuple = load_data(Compas())
    assert (6167, 400) == data.x.shape
    assert (6167, 1) == data.s.shape
    assert (6167, 1) == data.y.shape
    assert data.name == "Compas Sex"


def test_load_sqf():
    """test load sqf"""
    data: DataTuple = load_data(Sqf())
    assert (12347, 144) == data.x.shape
    assert (12347, 1) == data.s.shape
    assert (12347, 1) == data.y.shape
    assert data.name == "SQF Sex"


def test_load_german():
    """test load german"""
    data: DataTuple = load_data(German())
    assert (1000, 57) == data.x.shape
    assert (1000, 1) == data.s.shape
    assert (1000, 1) == data.y.shape
    assert data.name == "German Sex"


def test_load_german_ordered():
    """test load german ordered"""
    data: DataTuple = load_data(German(), ordered=True)
    assert (1000, 57) == data.x.shape
    assert (1000, 1) == data.s.shape
    assert (1000, 1) == data.y.shape


def test_load_adult_explicitly_sex():
    """test load adult explicitly sex"""
    data: DataTuple = load_data(Adult("Sex"))
    assert (45222, 101) == data.x.shape
    assert (45222, 1) == data.s.shape
    assert (45222, 1) == data.y.shape


def test_load_compas_explicitly_sex():
    """test load compas explicitly sex"""
    data: DataTuple = load_data(Compas("Sex"))
    assert (6167, 400) == data.x.shape
    assert (6167, 1) == data.s.shape
    assert (6167, 1) == data.y.shape


def test_load_compas_feature_length():
    """test load compas feature length"""
    data: DataTuple = load_data(Compas())
    assert len(Compas().ordered_features['x']) == 400
    assert len(Compas().discrete_features) == 395
    assert len(Compas().continuous_features) == 5
    assert data.s.shape == (6167, 1)
    assert data.y.shape == (6167, 1)


def test_load_credit_feature_length():
    """test load credit feature length"""
    data: DataTuple = load_data(Credit())
    assert len(Credit().ordered_features['x']) == 26
    assert len(Credit().discrete_features) == 6
    assert len(Credit().continuous_features) == 20
    assert data.s.shape == (30000, 1)
    assert data.y.shape == (30000, 1)


def test_load_adult_race():
    """test load adult race"""
    data: DataTuple = load_data(Adult("Race"))
    assert (45222, 98) == data.x.shape
    assert (45222, 5) == data.s.shape
    assert (45222, 1) == data.y.shape


def test_load_compas_race():
    """test load compas race"""
    data: DataTuple = load_data(Compas("Race"))
    assert (6167, 400) == data.x.shape
    assert (6167, 1) == data.s.shape
    assert (6167, 1) == data.y.shape


def test_load_adult_race_sex():
    """test load adult race sex"""
    data: DataTuple = load_data(Adult("Race-Sex"))
    assert (45222, 96) == data.x.shape
    assert (45222, 6) == data.s.shape
    assert (45222, 1) == data.y.shape


def test_load_compas_race_sex():
    """test load compas race sex"""
    data: DataTuple = load_data(Compas("Race-Sex"))
    assert (6167, 399) == data.x.shape
    assert (6167, 2) == data.s.shape
    assert (6167, 1) == data.y.shape


def test_load_adult_nationality():
    """test load adult nationality"""
    data: DataTuple = load_data(Adult("Nationality"))
    assert (45222, 62) == data.x.shape
    assert (45222, 41) == data.s.shape
    assert (45222, 1) == data.y.shape


def test_race_feature_split():
    """test race feature split"""
    adult: Adult = Adult(split="Custom")
    adult.sens_attrs = ["race_White"]
    adult.s_prefix = ["race"]
    adult.class_labels = ["salary_>50K"]
    adult.class_label_prefix = ["salary"]

    data: DataTuple = load_data(adult)

    assert (45222, 98) == data.x.shape
    assert (45222, 1) == data.s.shape
    assert (45222, 1) == data.y.shape


def test_additional_columns_load():
    """test additional columns load"""
    data_loc: Path = ROOT_PATH / "data" / "csvs" / "adult.csv"
    data_obj: Dataset = create_data_obj(
        data_loc,
        s_columns=["race_White"],
        y_columns=["salary_>50K"],
        additional_to_drop=["race_Black", "salary_<=50K"],
    )
    data: DataTuple = load_data(data_obj)

    assert (45222, 102) == data.x.shape
    assert (45222, 1) == data.s.shape
    assert (45222, 1) == data.y.shape


def test_domain_adapt_adult():
    """test domain adapt adult"""
    data: DataTuple = load_data(Adult())
    train, test = domain_split(
        datatup=data,
        tr_cond='education_Masters == 0. & education_Doctorate == 0.',
        te_cond='education_Masters == 1. | education_Doctorate == 1.',
    )
    assert (39106, 101) == train.x.shape
    assert (39106, 1) == train.s.shape
    assert (39106, 1) == train.y.shape

    assert (6116, 101) == test.x.shape
    assert (6116, 1) == test.s.shape
    assert (6116, 1) == test.y.shape

    data = load_data(Adult())
    train, test = domain_split(
        datatup=data, tr_cond='education_Masters == 0.', te_cond='education_Masters == 1.'
    )
    assert (40194, 101) == train.x.shape
    assert (40194, 1) == train.s.shape
    assert (40194, 1) == train.y.shape

    assert (5028, 101) == test.x.shape
    assert (5028, 1) == test.s.shape
    assert (5028, 1) == test.y.shape

    data = load_data(Adult())
    train, test = domain_split(
        datatup=data,
        tr_cond='education_Masters == 0. & education_Doctorate == 0. & education_Bachelors == 0.',
        te_cond='education_Masters == 1. | education_Doctorate == 1. | education_Bachelors == 1.',
    )
    assert (23966, 101) == train.x.shape
    assert (23966, 1) == train.s.shape
    assert (23966, 1) == train.y.shape

    assert (21256, 101) == test.x.shape
    assert (21256, 1) == test.s.shape
    assert (21256, 1) == test.y.shape


def test_query():
    """test query"""
    x: pd.DataFrame = pd.DataFrame(columns=['0a', 'b'], data=[[0, 1], [2, 3], [4, 5]])
    s: pd.DataFrame = pd.DataFrame(columns=['c='], data=[[6], [7], [8]])
    y: pd.DataFrame = pd.DataFrame(columns=['d'], data=[[9], [10], [11]])
    data = DataTuple(x=x, s=s, y=y, name='test_data')
    selected = query_dt(data, '_0a == 0 & c_eq_ == 6 & d == 9')
    pd.testing.assert_frame_equal(selected.x, x.head(1))
    pd.testing.assert_frame_equal(selected.s, s.head(1))
    pd.testing.assert_frame_equal(selected.y, y.head(1))


def test_concat():
    """test concat"""
    x: pd.DataFrame = pd.DataFrame(columns=['a'], data=[[1]])
    s: pd.DataFrame = pd.DataFrame(columns=['b'], data=[[2]])
    y: pd.DataFrame = pd.DataFrame(columns=['c'], data=[[3]])
    data1 = DataTuple(x=x, s=s, y=y, name='test_data')
    x = pd.DataFrame(columns=['a'], data=[[4]])
    s = pd.DataFrame(columns=['b'], data=[[5]])
    y = pd.DataFrame(columns=['c'], data=[[6]])
    data2 = DataTuple(x=x, s=s, y=y, name='test_tuple')
    data3 = concat_dt([data1, data2], axis='index', ignore_index=True)
    pd.testing.assert_frame_equal(data3.x, pd.DataFrame(columns=['a'], data=[[1], [4]]))
    pd.testing.assert_frame_equal(data3.s, pd.DataFrame(columns=['b'], data=[[2], [5]]))
    pd.testing.assert_frame_equal(data3.y, pd.DataFrame(columns=['c'], data=[[3], [6]]))



"""Algorithms by Zafar et al. for Demographic Parity."""
import json
from tempfile import TemporaryDirectory
from pathlib import Path
from typing import Dict, Any, Union, List
from abc import abstractmethod

import pandas as pd

from ethicml.algorithms.inprocess.installed_model import InstalledModel
from ethicml.utility.data_structures import DataTuple, TestTuple
from ethicml.preprocessing.adjust_labels import LabelBinarizer

SUB_DIR_IMPACT = Path(".") / "disparate_impact" / "run-classifier"
SUB_DIR_MISTREAT = Path(".") / "disparate_mistreatment" / "run_classifier"


class _ZafarAlgorithmBase(InstalledModel):
    def __init__(self, sub_dir: Path):
        super().__init__(
            dir_name="zafar",
            url="https://github.com/predictive-analytics-lab/fair-classification.git",
            top_dir="fair-classification",
        )
        self._sub_dir = sub_dir

    @staticmethod
    def _create_file_in_zafar_format(
        data: Union[DataTuple, TestTuple], file_path: Path, label_converter: LabelBinarizer
    ) -> None:
        """Save a DataTuple as a JSON file, which is extremely inefficient but what Zafar wants."""
        out: Dict[str, Any] = {}
        out["x"] = data.x.to_numpy().tolist()
        sens_attr = data.s.columns[0]
        out["sensitive"] = {}
        out["sensitive"][sens_attr] = data.s[sens_attr].to_numpy().tolist()
        if isinstance(data, DataTuple):
            data_converted = label_converter.adjust(data)
            class_attr = data.y.columns[0]
            out["class"] = (2 * data_converted.y[class_attr].to_numpy() - 1).tolist()
        else:
            out["class"] = [-1 for _ in range(data.x.shape[0])]
        with file_path.open("w") as out_file:
            json.dump(out, out_file)

    async def run_async(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        label_converter = LabelBinarizer()
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            train_path = tmp_path / "train.json"
            test_path = tmp_path / "test.json"
            self._create_file_in_zafar_format(train, train_path, label_converter)
            self._create_file_in_zafar_format(test, test_path, label_converter)
            predictions_path = tmp_path / "predictions.json"

            cmd = self._create_command_line(str(train_path), str(test_path), str(predictions_path))
            working_dir = self._code_path.resolve() / self._sub_dir
            await self._call_script(cmd, cwd=working_dir)
            predictions = predictions_path.open().read()
            predictions = json.loads(predictions)

        predictions_correct = pd.DataFrame([0 if x == -1 else 1 for x in predictions])
        return label_converter.post_only_labels(predictions_correct)

    @abstractmethod
    def _create_command_line(
        self, train_name: str, test_name: str, predictions_name: str
    ) -> List[str]:
        pass


class ZafarBaseline(_ZafarAlgorithmBase):
    """Zafar without fairness."""

    def __init__(self) -> None:
        """Init ZafarBaseline."""
        super().__init__(sub_dir=SUB_DIR_IMPACT)

    def _create_command_line(
        self, train_name: str, test_name: str, predictions_name: str
    ) -> List[str]:
        return ["main.py", train_name, test_name, predictions_name, 'baseline', '0']

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return "ZafarBaseline"


class ZafarAccuracy(_ZafarAlgorithmBase):
    """Zafar with fairness."""

    def __init__(self, gamma: float = 0.5):
        """Init ZafarAccuracy."""
        super().__init__(sub_dir=SUB_DIR_IMPACT)
        self.gamma = gamma

    def _create_command_line(
        self, train_name: str, test_name: str, predictions_name: str
    ) -> List[str]:
        return ["main.py", train_name, test_name, predictions_name, "gamma", str(self.gamma)]

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return f"ZafarAccuracy, γ={self.gamma}"


class ZafarFairness(_ZafarAlgorithmBase):
    """Zafar with fairness."""

    def __init__(self, c: float = 0.001):
        """Init ZafarFairness."""
        super().__init__(sub_dir=SUB_DIR_IMPACT)
        self._c = c

    def _create_command_line(
        self, train_name: str, test_name: str, predictions_name: str
    ) -> List[str]:
        return ["main.py", train_name, test_name, predictions_name, "c", str(self._c)]

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return f"ZafarFairness, c={self._c}"


class ZafarEqOpp(_ZafarAlgorithmBase):
    """Zafar for Equality of Opportunity."""

    _mode = "fnr"  # class level constant

    def __init__(self, tau: float = 5.0, mu: float = 1.2, eps: float = 0.0001):
        """Init Zafar."""
        super().__init__(sub_dir=SUB_DIR_MISTREAT)
        self._tau = tau
        self._mu = mu
        self._eps = eps

    def _create_command_line(
        self, train_name: str, test_name: str, predictions_name: str
    ) -> List[str]:
        return [
            "main.py",
            train_name,
            test_name,
            predictions_name,
            self._mode,
            str(self._tau),
            str(self._mu),
            str(self._eps),
        ]

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return f"ZafarEqOpp, τ={self._tau}, μ={self._mu}"


class ZafarEqOdds(ZafarEqOpp):
    """Zafar for Equalised Odds."""

    _mode = "fprfnr"

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return f"ZafarEqOdds, τ={self._tau}, μ={self._mu}"



"""Implementation for Louizos et al Variational Fair Autoencoder."""
# pylint: disable=arguments-differ

import torch
from torch import nn


class _OneHotEncoder(nn.Module):
    """One Hot Encode the output based on feature groups."""

    def __init__(self, n_dims, index_dim=1):
        super().__init__()
        self.n_dims = n_dims
        self.index_dim = index_dim

    def forward(self, x):
        indexes = x.argmax(dim=self.index_dim)
        indexes = indexes.type(torch.int64).view(-1, 1)
        n_dims = self.n_dims  # if self.n_dims is not None else int(torch.max(indexes)) + 1
        one_hots = torch.zeros(indexes.size()[0], n_dims).scatter_(1, indexes, 1)
        one_hots = one_hots.view(*indexes.shape, -1)
        return one_hots


class Categorical(nn.Module):
    """Group a category together."""

    def __init__(self, in_feat, dims):
        """Init Categorical object."""
        super().__init__()
        self.layer = nn.Sequential(nn.Linear(in_feat, dims), nn.Softmax(dim=-1))
        self.ohe = _OneHotEncoder(n_dims=dims)

    def forward(self, x):
        """Forward pass."""
        out = self.layer(x)
        if not self.training:
            out = self.ohe(out)
        return out



"""Base Class for decision / activation functions."""

from abc import ABC, abstractmethod
import numpy


class Activation(ABC):
    """Base class for decision functions."""

    @abstractmethod
    def apply(self, soft_output: numpy.ndarray) -> numpy.ndarray:
        """Apply the decision function to a soft prediction.

        Args:
            soft_output: soft prediction (i.e. a probability or logits)

        Returns:
            decision
        """

    @abstractmethod
    def get_name(self) -> str:
        """Name of activation function."""



"""Simply returns the majority label from the train set."""

import pandas as pd

from ethicml.common import implements
from ethicml.algorithms.inprocess.in_algorithm import InAlgorithm
from ethicml.utility.data_structures import DataTuple, TestTuple


class Majority(InAlgorithm):
    """Simply returns the majority label from the train set."""

    @implements(InAlgorithm)
    def run(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        maj = train.y.mode().to_numpy()
        return pd.DataFrame(maj.repeat(len(test.x)), columns=["preds"])

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return "Majority"



"""Class for loading CelebA.

Modifies the Pytorch CelebA dataset by enabling the use of sensitive attributes
and biased subset sampling.
"""
import os
import warnings
from pathlib import Path
from typing import Tuple, Optional, Callable, Sequence
from typing_extensions import Literal

import pandas as pd
from PIL import Image

import torch
from torch import Tensor
from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import check_integrity, download_file_from_google_drive

from ethicml.preprocessing import get_biased_subset, SequentialSplit
from ethicml.utility import DataTuple

__all__ = ["CelebA"]


_CELEBATTRS = Literal[
    "5_o_Clock_Shadow",
    "Arched_Eyebrows",
    "Attractive",
    "Bags_Under_Eyes",
    "Bald",
    "Bangs",
    "Big_Lips",
    "Big_Nose",
    "Black_Hair",
    "Blond_Hair",
    "Blurry",
    "Brown_Hair",
    "Bushy_Eyebrows",
    "Chubby",
    "Double_Chin",
    "Eyeglasses",
    "Goatee",
    "Gray_Hair",
    "Heavy_Makeup",
    "High_Cheekbones",
    "Male",
    "Mouth_Slightly_Open",
    "Mustache",
    "Narrow_Eyes",
    "No_Beard",
    "Oval_Face",
    "Pale_Skin",
    "Pointy_Nose",
    "Receding_Hairline",
    "Rosy_Cheeks",
    "Sideburns",
    "Smiling",
    "Straight_Hair",
    "Wavy_Hair",
    "Wearing_Earrings",
    "Wearing_Hat",
    "Wearing_Lipstick",
    "Wearing_Necklace",
    "Wearing_Necktie",
    "Young",
]


class CelebA(VisionDataset):
    """Large-scale CelebFaces Attributes (CelebA) Dataset."""

    base_folder = "celeba"

    file_list = [
        (
            "0B7EVK8r0v71pZjFTYXZWM3FlRnM",  # File ID
            "00d2c5bc6d35e252742224ab0c1e8fcb",  # MD5 Hash
            "img_align_celeba.zip",  # Filename
        ),
        (
            "0B7EVK8r0v71pblRyaVFSWGxPY0U",
            "75e246fa4810816ffd6ee81facbd244c",
            "list_attr_celeba.txt",
        ),
        (
            "0B7EVK8r0v71pY0NSMzRuSXJEVkk",
            "d32c9cbf5e040fd4025c592c306e6668",
            "list_eval_partition.txt",
        ),
    ]

    def __init__(
        self,
        root: str,
        biased: bool,
        mixing_factor: float,
        unbiased_pcnt: float,
        sens_attrs: Sequence[_CELEBATTRS],
        target_attr_name: _CELEBATTRS,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
        seed: int = 42,
    ):
        """Large-scale CelebFaces Attributes (CelebA) Dataset.

        <http://mmlab.ie.cuhk.edu.hk/projects/CelebA.html>
        Adapted from torchvision.datasets to enable the loading of data triplets and biased/unbiased
        subsets while removing superfluous (for our purposes) elements of the dataset (e.g. facial
        landmarks).

        Args:
            root: Root directory where images are downloaded to.
            biased: Wheher to artifically bias the dataset according to the mixing factor. See
                    :func:`get_biased_subset()` for more details.
            mixing_factor: Mixing factor used to generate the biased subset of the data.
            sens_attrs: Attribute(s) to set as the sensitive attribute. Biased sampling cannot be
                        performed if multiple sensitive attributes are specified.
            unbiased_pcnt: Percentage of the dataset to set aside as the 'unbiased' split.
            target_attr_name: Attribute to set as the target attribute.
            transform: A function/transform that  takes in an PIL image and returns a transformed
                       version. E.g, `transforms.ToTensor`
            target_transform: A function/transform that takes in the target and transforms it.
            download: If true, downloads the dataset from the internet and puts it in root
                      directory. If dataset is already downloaded, it is not downloaded again.
            seed: Random seed used to sample biased subset.
        """
        super().__init__(root, transform=transform, target_transform=target_transform)

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted." + " You can use download=True to download it"
            )

        base = Path(self.root) / self.base_folder
        partition_file = base / "list_eval_partition.txt"
        # partition: information about which samples belong to train, val or test
        partition = pd.read_csv(
            partition_file, delim_whitespace=True, header=None, index_col=0, names=["partition"]
        )
        # attrs: all attributes with filenames as index
        attrs = pd.read_csv(base / "list_attr_celeba.txt", delim_whitespace=True, header=1)
        all_data = pd.concat([partition, attrs], axis="columns", sort=False)
        # the filenames are used for indexing; here we turn them into a regular column
        all_data = all_data.reset_index(drop=False).rename(columns={"index": "filenames"})

        attr_names = list(attrs.columns)

        # Multiple attributes have been designated as sensitive
        # Note that in this case biased dataset sampling cannot be performed
        if len(sens_attrs) > 1:
            if any(sens_attr_name not in attr_names for sens_attr_name in sens_attrs):
                raise ValueError(f"at least one of {sens_attrs} does not exist as an attribute.")
            # only use those samples where exactly one of the specified attributes is true
            all_data = all_data.loc[((all_data[sens_attrs] + 1) // 2).sum(axis="columns") == 1]
            self.s_dim = len(sens_attrs)
            # perform the reverse operation of one-hot encoding
            data_only_sens = all_data[sens_attrs]
            data_only_sens.columns = list(range(self.s_dim))
            sens_attr = data_only_sens.idxmax(axis="columns").to_frame(name=",".join(sens_attrs))
        else:
            sens_attr_name = sens_attrs[0].capitalize()
            if sens_attr_name not in attr_names:
                raise ValueError(f"{sens_attr_name} does not exist as an attribute.")
            sens_attr = all_data[[sens_attr_name]]
            sens_attr = (sens_attr + 1) // 2  # map from {-1, 1} to {0, 1}
            self.s_dim = 1

        target_attr_name = target_attr_name.capitalize()
        if target_attr_name not in attr_names:
            raise ValueError(f"{target_attr_name} does not exist as an attribute.")

        if sens_attr_name == target_attr_name:
            raise ValueError(f"{sens_attr_name} does not exist as an attribute.")

        if sens_attr_name == target_attr_name:
            warnings.warn("Same attribute specified for both the sensitive and target attribute.")

        target_attr = all_data[[target_attr_name]]
        target_attr: pd.DataFrame = (target_attr + 1) // 2  # map from {-1, 1} to {0, 1}

        filename = all_data[["filenames"]]

        all_dt = DataTuple(x=filename, s=sens_attr, y=target_attr)

        # NOTE: the sequential split does not shuffle
        unbiased_dt, biased_dt, _ = SequentialSplit(train_percentage=unbiased_pcnt)(all_dt)

        if biased:
            if self.s_dim == 1:  # FIXME: biasing the dataset only works with binary s right now
                biased_dt, _ = get_biased_subset(
                    data=biased_dt, mixing_factor=mixing_factor, unbiased_pcnt=0, seed=seed
                )
            filename, sens_attr, target_attr = biased_dt
        else:
            filename, sens_attr, target_attr = unbiased_dt

        self.filename = filename.to_numpy()[:, 0]
        self.sens_attr = torch.as_tensor(sens_attr.to_numpy())
        self.target_attr = torch.as_tensor(target_attr.to_numpy())

    def _check_integrity(self) -> bool:
        """Check integrity of the data folder.

        Returns:
            bool: Boolean indicating whether the file containing the celeba data
                  is readable.
        """
        base = Path(self.root) / self.base_folder
        for (_, md5, filename) in self.file_list:
            fpath = base / filename
            ext = fpath.suffix
            # Allow original archive to be deleted (zip and 7z)
            # Only need the extracted images
            if ext not in [".zip", ".7z"] and not check_integrity(str(fpath), md5):
                return False

        # Should check a hash of the images
        return (base / "img_align_celeba").is_dir()

    def download(self) -> None:
        """Attempt to download data if files cannot be found in the base folder."""
        import zipfile

        if self._check_integrity():
            print("Files already downloaded and verified")
            return

        for (file_id, md5, filename) in self.file_list:
            download_file_from_google_drive(
                file_id, os.path.join(self.root, self.base_folder), filename, md5
            )

        with zipfile.ZipFile(
            os.path.join(self.root, self.base_folder, "img_align_celeba.zip"), "r"
        ) as fhandle:
            fhandle.extractall(os.path.join(self.root, self.base_folder))

    def __getitem__(self, index: int) -> Tuple[Tensor, Tensor, Tensor]:
        """Fetch the data sample at the given index.

        Args:
            index (int): Index of the sample to be loaded in.

        Returns:
            Tuple[1]: Tuple containing the sample along
            with its sensitive and target attribute labels.
        """
        x = Image.open(
            os.path.join(self.root, self.base_folder, "img_align_celeba", self.filename[index])
        )
        s = self.sens_attr[index]
        target = self.target_attr[index]

        if self.transform is not None:
            x = self.transform(x)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return x, s, target

    def __len__(self) -> int:
        """Length (sample count) of the dataset.

        Returns:
            Integer indicating the length of the dataset.
        """
        return len(self.sens_attr)



"""Simple upsampler that makes subgroups the same size as the majority group."""
import itertools
from typing import List, Tuple, Dict, Optional
from typing_extensions import Literal

import pandas as pd

from ethicml.common import implements
from ethicml.algorithms.inprocess import LRProb
from ethicml.utility import TestTuple, DataTuple
from .pre_algorithm import PreAlgorithm


class Upsampler(PreAlgorithm):
    """Upsampler algorithm.

    Given a datatuple, create a larger datatuple such that the subgroups have a balanced number
    of samples.
    """

    def __init__(self, strategy: Literal["uniform", "preferential", "naive"] = "uniform"):
        """Init Upsampler."""
        super().__init__()

        assert strategy in ["uniform", "preferential", "naive"]
        self.strategy = strategy

    @implements(PreAlgorithm)
    def run(self, train: DataTuple, test: TestTuple) -> Tuple[DataTuple, TestTuple]:
        return upsample(train, test, self.strategy)

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return f"Upsample {self.strategy}"


def concat_datatuples(first_dt: DataTuple, second_dt: DataTuple) -> DataTuple:
    """Given 2 datatuples, concatenate them and shuffle."""
    assert (first_dt.x.columns == second_dt.x.columns).all()
    assert (first_dt.s.columns == second_dt.s.columns).all()
    assert (first_dt.y.columns == second_dt.y.columns).all()

    x_columns: pd.Index = first_dt.x.columns
    s_columns: pd.Index = first_dt.s.columns
    y_columns: pd.Index = first_dt.y.columns

    a_combined: pd.DataFrame = pd.concat([first_dt.x, first_dt.s, first_dt.y], axis="columns")
    b_combined: pd.DataFrame = pd.concat([second_dt.x, second_dt.s, second_dt.y], axis="columns")

    combined: pd.DataFrame = pd.concat([a_combined, b_combined], axis="index")
    combined = combined.sample(frac=1.0, random_state=1).reset_index(drop=True)

    return DataTuple(
        x=combined[x_columns], s=combined[s_columns], y=combined[y_columns], name=first_dt.name
    )


def upsample(
    dataset: DataTuple, test: TestTuple, strategy: Literal["uniform", "preferential", "naive"]
) -> Tuple[DataTuple, TestTuple]:
    """Upsample a datatuple."""
    s_col = dataset.s.columns[0]
    y_col = dataset.y.columns[0]

    s_vals: List[int] = list(map(int, dataset.s[s_col].unique()))
    y_vals: List[int] = list(map(int, dataset.y[y_col].unique()))

    groups = itertools.product(s_vals, y_vals)

    data: Dict[Tuple[int, int], DataTuple] = {}
    for s, y in groups:
        s_y_mask = (dataset.s[s_col] == s) & (dataset.y[y_col] == y)
        data[(s, y)] = DataTuple(
            x=dataset.x.loc[s_y_mask].reset_index(drop=True),
            s=dataset.s.loc[s_y_mask].reset_index(drop=True),
            y=dataset.y.loc[s_y_mask].reset_index(drop=True),
            name=dataset.name,
        )

    percentages: Dict[Tuple[int, int], float] = {}

    vals: List[int] = []
    for key, val in data.items():
        vals.append(val.x.shape[0])

    for key, val in data.items():
        if strategy == "naive":
            percentages[key] = max(vals) / val.x.shape[0]
        else:
            s_val: int = key[0]
            y_val: int = key[1]

            y_eq_y = dataset.y.loc[dataset.y[y_col] == y_val].count().to_numpy()[0]
            s_eq_s = dataset.s.loc[dataset.s[s_col] == s_val].count().to_numpy()[0]

            num_samples = dataset.y.count().to_numpy()[0]
            num_batch = val.y.count().to_numpy()[0]

            percentages[key] = round((y_eq_y * s_eq_s / (num_batch * num_samples)), 8)

    x_columns: pd.Index = dataset.x.columns
    s_columns: pd.Index = dataset.s.columns
    y_columns: pd.Index = dataset.y.columns

    upsampled: Dict[Tuple[int, int], DataTuple] = {}
    for key, val in data.items():
        all_data: pd.DataFrame = pd.concat([val.x, val.s, val.y], axis="columns")
        all_data = all_data.sample(frac=percentages[key], random_state=1, replace=True).reset_index(
            drop=True
        )
        upsampled[key] = DataTuple(
            x=all_data[x_columns], s=all_data[s_columns], y=all_data[y_columns], name=dataset.name
        )

    upsampled_datatuple: Optional[DataTuple] = None
    for key, val in upsampled.items():
        if upsampled_datatuple is None:
            upsampled_datatuple = val
        else:
            upsampled_datatuple = concat_datatuples(upsampled_datatuple, val)

    if strategy == "preferential":
        ranker = LRProb()
        rank = ranker.run(dataset, dataset)

        selected: List[pd.DataFrame] = []

        all_data = pd.concat([dataset.x, dataset.s, dataset.y], axis="columns")
        all_data = pd.concat([all_data, rank], axis="columns")

        for key, val in data.items():

            s_val = key[0]
            y_val = key[1]
            s_y_mask = (dataset.s[s_col] == s_val) & (dataset.y[y_col] == y_val)

            ascending = False
            if s_val <= 0:
                ascending = True

            if percentages[key] > 1.0:
                selected.append(all_data.loc[s_y_mask])
                percentages[key] -= 1.0

            weight = all_data.loc[s_y_mask][y_col].count()
            selected.append(
                all_data.loc[s_y_mask]
                .sort_values(by=['preds'], ascending=ascending)
                .iloc[: int(percentages[key] * weight)]
            )

        upsampled_dataframes: pd.DataFrame
        for i, df in enumerate(selected):
            if i == 0:
                upsampled_dataframes = df.drop(['preds'], axis="columns")
            else:
                upsampled_dataframes = pd.concat(
                    [upsampled_dataframes, df.drop(['preds'], axis="columns")], axis="index"
                ).reset_index(drop=True)
        upsampled_datatuple = DataTuple(
            x=upsampled_dataframes[x_columns],
            s=upsampled_dataframes[s_columns],
            y=upsampled_dataframes[y_columns],
            name=dataset.name,
        )

    assert upsampled_datatuple is not None
    return upsampled_datatuple, TestTuple(x=test.x, s=test.s, name=test.name)



"""Create plots of a dataset."""
from pathlib import Path
from typing import Tuple, List, Union, Optional
import itertools

from typing_extensions import Literal
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib as mpl

from ethicml.metrics import Metric
from ethicml.utility.data_structures import DataTuple, Results
from ethicml.visualisation.common import errorbox, DataEntry, PlotDef, LegendType, PlotType, scatter

MARKERS = ["s", "p", "P", "*", "+", "x", "o", "v"]


def save_2d_plot(data: DataTuple, filepath: str) -> None:
    """Make 2D plot."""
    file_path = Path(filepath)
    columns = data.x.columns

    amalgamated = pd.concat([data.x, data.s, data.y], axis="columns")

    plot = sns.scatterplot(
        x=columns[0],
        y=columns[1],
        hue=data.y.columns[0],
        palette="Set2",
        data=amalgamated,
        style=data.s.columns[0],
        legend="full",
    )

    file_path.parent.mkdir(exist_ok=True)
    plot.figure.savefig(file_path)
    plt.clf()


def save_jointplot(data: DataTuple, filepath: str, dims: Tuple[int, int] = (0, 1)) -> None:
    """Make joint plot."""
    file_path = Path(filepath)
    columns = data.x.columns

    amalgamated = pd.concat([data.x, data.y], axis="columns")

    plot = sns.jointplot(x=columns[dims[0]], y=columns[dims[1]], data=amalgamated, kind="kde")

    file_path.parent.mkdir(exist_ok=True)
    plot.savefig(file_path)
    plt.clf()


def make_gif(files: List[str], name: str = "movie") -> None:
    """Make GIF."""
    import imageio

    images = []
    for filename in files:
        images.append(imageio.imread(filename))
    imageio.mimsave(f"{name}.gif", images)


def save_label_plot(data: DataTuple, filename: str) -> None:
    """Make label plot."""
    file_path = Path(filename)

    # Only consider 1 sens attr for now
    s_col = data.s.columns[0]
    s_values = data.s[s_col].value_counts() / data.s[s_col].count()

    s_0_val = s_values[0]
    s_1_val = s_values[1]

    s_0_label = s_values.index[0]
    s_1_label = s_values.index[1]

    y_col = data.y.columns[0]
    y_s0 = (
        data.y[y_col][data.s[s_col] == 0].value_counts() / data.y[y_col][data.s[s_col] == 0].count()
    )
    y_s1 = (
        data.y[y_col][data.s[s_col] == 1].value_counts() / data.y[y_col][data.s[s_col] == 1].count()
    )

    y_0_label = y_s0.index[0]
    y_1_label = y_s0.index[1]

    mpl.style.use("seaborn-pastel")
    # plt.xkcd()

    fig, plot = plt.subplots()

    quadrant1 = plot.bar(
        0,
        height=y_s0[y_0_label] * 100,
        width=s_0_val * 100,
        align="edge",
        edgecolor="black",
        color="C0",
    )
    quadrant2 = plot.bar(
        s_0_val * 100,
        height=y_s1[y_0_label] * 100,
        width=s_1_val * 100,
        align="edge",
        edgecolor="black",
        color="C1",
    )
    quadrant3 = plot.bar(
        0,
        height=y_s0[y_1_label] * 100,
        width=s_0_val * 100,
        bottom=y_s0[y_0_label] * 100,
        align="edge",
        edgecolor="black",
        color="C2",
    )
    quadrant4 = plot.bar(
        s_0_val * 100,
        height=y_s1[y_1_label] * 100,
        width=s_1_val * 100,
        bottom=y_s1[y_0_label] * 100,
        align="edge",
        edgecolor="black",
        color="C3",
    )

    plot.set_ylim(0, 100)
    plot.set_xlim(0, 100)
    plot.set_ylabel(f"Percent {y_col}=y")
    plot.set_xlabel(f"Percent {s_col}=s")
    plot.set_title("Dataset Composition by class and sensitive attribute")

    plot.legend(
        [quadrant1, quadrant2, quadrant3, quadrant4],
        [
            f"y={y_0_label}, s={s_0_label}",
            f"y={y_0_label}, s={s_1_label}",
            f"y={y_1_label}, s={s_0_label}",
            f"y={y_1_label}, s={s_1_label}",
        ],
    )

    file_path.parent.mkdir(exist_ok=True)
    fig.savefig(file_path)


def single_plot(
    plot: plt.Axes,
    results: Results,
    xaxis: Tuple[str, str],
    yaxis: Tuple[str, str],
    dataset: str,
    transform: str,
    ptype: PlotType = "box",
    legend_pos: Optional[LegendType] = "outside",
    legend_yanchor: float = 1.0,
    markersize: int = 6,
) -> Union[None, Literal[False], mpl.legend.Legend]:
    """Provide the functionality of the individual plotting functions through a nice interface.

    This function can also be used to create figures with multiple plots on them, because it does
    not generate a Figure object itself.

    Args:
        plot: Plot object
        results: DataFrame with the data
        xaxis: name of column that's plotted on the x-axis
        yaxis: name of column that's plotted on the y-axis
        dataset: string that identifies the dataset
        transform: string that identifies the preprocessing method
        ptype: plot type
        legend_pos: position of the legend (or None for no legend)
        legend_yanchor: position in the vertical direction where the legend should begin
        markersize: size of marker

    Returns:
        the legend object if something was plotted; False otherwise
    """
    results_df = results.data
    mask_for_dataset = results_df.index.get_level_values('dataset') == dataset
    mask_for_transform = results_df.index.get_level_values('transform') == transform
    matching_results = results_df.loc[mask_for_dataset & mask_for_transform]

    if pd.isnull(matching_results[[xaxis[0], yaxis[0]]]).any().any():
        return False  # nothing to plot

    entries: List[DataEntry] = []
    for count, model in enumerate(results_df.index.to_frame()['model'].unique()):
        mask_for_model = results_df.index.get_level_values('model') == model
        data = results_df.loc[mask_for_dataset & mask_for_model & mask_for_transform]
        model_label = f"{model} ({transform})" if transform != "no_transform" else str(model)
        entries.append(DataEntry(model_label, data, count % 2 == 0))

    title = f"{dataset}, {transform}"
    plot_def = PlotDef(
        title=title, entries=entries, legend_pos=legend_pos, legend_yanchor=legend_yanchor
    )
    if ptype == "box":
        return errorbox(plot, plot_def, xaxis, yaxis, 0, 0, markersize, use_cross=False)
    if ptype == "cross":
        return errorbox(plot, plot_def, xaxis, yaxis, 0, 0, markersize, use_cross=True)
    if ptype == "scatter":
        return scatter(plot, plot_def, xaxis, yaxis, 0, markersize, connect_dots=False)
    if ptype == "line":
        return scatter(plot, plot_def, xaxis, yaxis, 0, markersize, connect_dots=True)
    raise ValueError(f"Unknown plot type '{ptype}'")


def plot_results(
    results: Results,
    metric_y: Union[str, Metric],
    metric_x: Union[str, Metric],
    ptype: PlotType = "box",
    save: bool = True,
    dpi: int = 300,
) -> List[Tuple[plt.Figure, plt.Axes]]:
    """Plot the given result with boxes that represent mean and standard deviation.

    Args:
        results: a DataFrame that already contains the values of the metrics
        metric_y: a Metric object or a column name that defines which metric to plot on the y-axis
        metric_x: a Metric object or a column name that defines which metric to plot on the x-axis
        ptype: plot type
        save: if True, save the plot as a PDF
        dpi: DPI of the plots
        use_cross: if True, use crosses with error bars instead of boxes

    Returns:
        A list of all figures and plots
    """
    directory = Path(".") / "plots"
    directory.mkdir(exist_ok=True)
    results_df = results.data

    def _get_columns(metric: Metric) -> List[str]:
        cols = [col for col in results_df.columns if metric.name in col]
        if not cols:
            raise ValueError(f"No matching columns found for Metric \"{metric.name}\".")
        # if there are multiple matches, then the metric was `per_sensitive_attribute`. In this
        # case, we *only* want ratios and differences; not the plain result
        if len(cols) > 1:
            cols = [col for col in cols if ("-" in col) or ("/" in col)]
        return cols

    if isinstance(metric_x, Metric):
        # if the metric is given as a Metric object, look for matching columns
        cols_x = _get_columns(metric_x)
    else:
        if metric_x not in results_df.columns:
            raise ValueError(f"No column named \"{metric_x}\".")
        cols_x = [metric_x]

    if isinstance(metric_y, Metric):
        cols_y = _get_columns(metric_y)
    else:
        if metric_y not in results_df.columns:
            raise ValueError(f"No column named \"{metric_y}\".")
        cols_y = [metric_y]

    # generate the Cartesian product of `cols_x` and `cols_y`; i.e. all possible combinations
    # this preserves the order of x and y
    possible_pairs = list(itertools.product(cols_x, cols_y))

    figure_list: List[Tuple[plt.Figure, plt.Axes]] = []
    for dataset in results_df.index.to_frame()['dataset'].unique():
        dataset_: str = str(dataset)
        for transform in results_df.index.to_frame()['transform'].unique():
            transform_: str = str(transform)
            for x_axis, y_axis in possible_pairs:
                fig: plt.Figure
                plot: plt.Axes
                fig, plot = plt.subplots(dpi=dpi)

                xtuple = (x_axis, x_axis.replace("_", " "))
                ytuple = (y_axis, y_axis.replace("_", " "))
                legend = single_plot(
                    plot, results, xtuple, ytuple, dataset_, transform_, ptype=ptype
                )

                if legend is False:  # use "is" here because otherwise any falsy value would match
                    plt.close(fig)
                    continue  # nothing was plotted -> don't save it and don't add it to the list

                if save:
                    metric_a = x_axis.replace("/", "_over_")
                    metric_b = y_axis.replace("/", "_over_")
                    fig.savefig(
                        directory / f"{dataset} {transform} {metric_a} {metric_b}.pdf",
                        bbox_inches='tight',
                    )
                plt.close(fig)
                figure_list += [(fig, plot)]
    return figure_list



"""Base class for Algorithms."""
import sys
from abc import ABC, abstractmethod, ABCMeta
from typing import List, Optional, Dict, Coroutine, TypeVar, Any
import asyncio
from pathlib import Path


class Algorithm(ABC):
    """Base class for Algorithms."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the algorithm."""


class AlgorithmAsync(metaclass=ABCMeta):
    """Base class of async methods; meant to be used in conjuction with :class:`Algorithm`."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the algorithm."""

    @property
    def _executable(self) -> str:
        """Path to a (Python) executable.

        By default, the Python executable that called this script is used.
        """
        return sys.executable

    async def _call_script(
        self, cmd_args: List[str], env: Optional[Dict[str, str]] = None, cwd: Optional[Path] = None
    ) -> None:
        """This function calls a (Python) script as a separate process.

        An exception is thrown if the called script failed.

        Args:
            cmd_args: list of strings that are passed as commandline arguments to the executable
            env: environment variables specified as a dictionary; e.g. {"PATH": "/usr/bin"}
            cwd: if not None, change working directory to the given path before running command
        """
        process = await asyncio.create_subprocess_exec(  # wait for process creation to finish
            self._executable,
            *cmd_args,
            stderr=asyncio.subprocess.PIPE,  # we capture the stderr for errors
            env=env,
            cwd=cwd,
        )
        # print(f"Started: {cmd_args!r} (pid = {process.pid})")

        try:  # wait for process itself to finish
            one_hour = 3600
            _, stderr = await asyncio.wait_for(process.communicate(), one_hour)
        except asyncio.TimeoutError as error:
            raise RuntimeError("The script timed out.") from error

        if process.returncode != 0:
            print(f"Failure: {cmd_args!r} (pid = {process.pid})")
            if stderr:
                print(stderr.decode().strip())
            raise RuntimeError(
                f"The script failed. Supplied arguments: {cmd_args} with exec: {self._executable}"
            )


_T = TypeVar("_T")


def run_blocking(promise: Coroutine[Any, Any, _T]) -> _T:
    """Run an asynchronous process as a blocking process."""
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(promise)



"""
Test the saving data capability
"""
from pathlib import Path
from tempfile import TemporaryDirectory

import pandas as pd
import numpy as np

import pytest

from ethicml.utility import DataTuple, TestTuple
from ethicml.algorithms import run_blocking
from ethicml.algorithms.inprocess import InAlgorithmAsync


def test_simple_saving() -> None:
    """
    tests that a DataTuple can be saved
    """
    data_tuple = DataTuple(
        x=pd.DataFrame({'a1': np.array([3.2, 9.4, np.nan, 0.0]), 'a2': np.array([1, 1, 0, 1])}),
        s=pd.DataFrame(
            {
                'b1': np.array([1.8, -0.3, 1e10]),
                'b2': np.array([1, 1, -1]),
                'b3': np.array([0, 1, 0]),
            }
        ),
        y=pd.DataFrame({'c1': np.array([-2, -3, np.nan]), 'c3': np.array([0, 1, 0])}),
        name='test data',
    )

    class CheckEquality(InAlgorithmAsync):
        """Dummy algorithm class for testing whether writing and reading feather files works"""

        @property
        def name(self) -> str:
            return "Check equality"

        def _script_command(self, train_paths, _, pred_path):
            """Check if the dataframes loaded from the files are the same as the original ones"""
            loaded = train_paths.load_from_feather()
            pd.testing.assert_frame_equal(data_tuple.x, loaded.x)
            pd.testing.assert_frame_equal(data_tuple.s, loaded.s)
            pd.testing.assert_frame_equal(data_tuple.y, loaded.y)
            # the following command copies the x of the training data to the pred_path location
            return ['-c', f'import shutil; shutil.copy("{train_paths.x}", "{pred_path}")']

    data_x = run_blocking(CheckEquality().run_async(data_tuple, data_tuple))
    pd.testing.assert_frame_equal(data_tuple.x, data_x)


def test_dataset_name_none() -> None:
    """
    tests that a DataTuple can be saved without the name property
    """
    datatup = DataTuple(
        x=pd.DataFrame([3.0], columns=['a1']),
        s=pd.DataFrame([4.0], columns=['b2']),
        y=pd.DataFrame([6.0], columns=['c3']),
        name=None,
    )
    with TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)
        paths = datatup.write_as_feather(tmp_path, "pytest")
        # reload from feather file
        reloaded = paths.load_from_feather()
    assert reloaded.name is None
    pd.testing.assert_frame_equal(datatup.x, reloaded.x)
    pd.testing.assert_frame_equal(datatup.s, reloaded.s)
    pd.testing.assert_frame_equal(datatup.y, reloaded.y)


def test_dataset_name_with_spaces() -> None:
    """
    tests that a dataset name can contain spaces and special chars
    """
    name = "This is a very@#$%^&*((())) complicated name"
    datatup = TestTuple(
        x=pd.DataFrame([3.0], columns=['a1']), s=pd.DataFrame([4.0], columns=['b2']), name=name
    )
    with TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)
        paths = datatup.write_as_feather(tmp_path, "pytest2")
        # reload from feather file
        reloaded = paths.load_from_feather()
    assert name == reloaded.name
    pd.testing.assert_frame_equal(datatup.x, reloaded.x)
    pd.testing.assert_frame_equal(datatup.s, reloaded.s)


def test_apply_to_joined_df() -> None:
    """
    tests apply_to_joined_df_function
    """
    datatup = DataTuple(
        x=pd.DataFrame([3.0], columns=['a1']),
        s=pd.DataFrame([4.0], columns=['b2']),
        y=pd.DataFrame([6.0], columns=['c3']),
        name=None,
    )

    def _identity(x: pd.DataFrame):
        return x

    result = datatup.apply_to_joined_df(_identity)
    pd.testing.assert_frame_equal(datatup.x, result.x)
    pd.testing.assert_frame_equal(datatup.s, result.s)
    pd.testing.assert_frame_equal(datatup.y, result.y)


def test_data_tuple_len() -> None:
    """test DataTuple len property"""
    datatup_unequal_len = DataTuple(
        x=pd.DataFrame([3.0, 2.0], columns=['a1']),
        s=pd.DataFrame([4.0], columns=['b2']),
        y=pd.DataFrame([6.0], columns=['c3']),
        name=None,
    )
    with pytest.raises(AssertionError):
        len(datatup_unequal_len)

    datatup_equal_len = DataTuple(
        x=pd.DataFrame([3.0, 2.0, 1.0], columns=['a1']),
        s=pd.DataFrame([4.0, 5.0, 9.0], columns=['b2']),
        y=pd.DataFrame([6.0, 4.2, 6.7], columns=['c3']),
        name=None,
    )
    assert len(datatup_equal_len) == 3



"""
Test preprocessing capabilities
"""
import math
from typing import Tuple

from pytest import approx
import pandas as pd

from ethicml.utility import DataTuple
from ethicml.data import load_data, Toy, Adult
from ethicml.preprocessing import (
    ProportionalSplit,
    SequentialSplit,
    train_test_split,
    bin_cont_feats,
    get_biased_subset,
    get_biased_and_debiased_subsets,
)

from .run_algorithm_test import count_true


def test_train_test_split():
    """test train test split"""
    data: DataTuple = load_data(Toy())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    assert train is not None
    assert test is not None
    assert train.x.shape[0] > test.x.shape[0]
    assert train.x['a1'].values[0] == 2.7839204628851526
    assert train.x['a2'].values[0] == 6.083679468491366

    assert train.x.shape[0] == train.s.shape[0]
    assert train.s.shape[0] == train.y.shape[0]

    len_default = math.floor((2000 / 100) * 80)
    assert train.s.shape[0] == len_default
    assert test.s.shape[0] == 2000 - len_default

    len_0_9 = math.floor((2000 / 100) * 90)
    train, test = train_test_split(data, train_percentage=0.9)
    assert train.s.shape[0] == len_0_9
    assert test.s.shape[0] == 2000 - len_0_9

    len_0_7 = math.floor((2000 / 100) * 70)
    train, test = train_test_split(data, train_percentage=0.7)
    assert train.s.shape[0] == len_0_7
    assert test.s.shape[0] == 2000 - len_0_7

    len_0_5 = math.floor((2000 / 100) * 50)
    train, test = train_test_split(data, train_percentage=0.5)
    assert train.s.shape[0] == len_0_5
    assert test.s.shape[0] == 2000 - len_0_5

    len_0_3 = math.floor((2000 / 100) * 30)
    train, test = train_test_split(data, train_percentage=0.3)
    assert train.s.shape[0] == len_0_3
    assert test.s.shape[0] == 2000 - len_0_3

    len_0_1 = math.floor((2000 / 100) * 10)
    train, test = train_test_split(data, train_percentage=0.1)
    assert train.s.shape[0] == len_0_1
    assert test.s.shape[0] == 2000 - len_0_1

    len_0_0 = math.floor((2000 / 100) * 0)
    train, test = train_test_split(data, train_percentage=0.0)
    assert train.s.shape[0] == len_0_0
    assert train.name == "Toy - Train"
    assert test.s.shape[0] == 2000 - len_0_0
    assert test.name == "Toy - Test"


def test_prop_train_test_split():
    """test prop train test split"""
    data: DataTuple = load_data(Toy())
    train: DataTuple
    test: DataTuple
    train, test, _ = ProportionalSplit(train_percentage=0.8)(data, split_id=0)
    assert train is not None
    assert test is not None
    assert train.x.shape[0] > test.x.shape[0]
    assert train.x['a1'].values[0] == 2.9624320112634743
    assert train.x['a2'].values[0] == 1.996089217523236

    assert train.x.shape[0] == train.s.shape[0]
    assert train.s.shape[0] == train.y.shape[0]

    len_default = math.floor((2000 / 100) * 80)
    assert train.s.shape[0] == len_default
    assert test.s.shape[0] == 2000 - len_default

    # assert that the proportion of s=0 to s=1 has remained the same (and also test for y=0/y=1)
    assert count_true(train.s.to_numpy() == 0) == round(0.8 * count_true(data.s.to_numpy() == 0))
    assert count_true(train.y.to_numpy() == 0) == round(0.8 * count_true(data.y.to_numpy() == 0))

    len_0_9 = math.floor((2000 / 100) * 90)
    train, test, _ = ProportionalSplit(train_percentage=0.9)(data, split_id=0)
    assert train.s.shape[0] == len_0_9
    assert test.s.shape[0] == 2000 - len_0_9
    assert count_true(train.s.to_numpy() == 0) == round(0.9 * count_true(data.s.to_numpy() == 0))
    assert count_true(train.y.to_numpy() == 0) == round(0.9 * count_true(data.y.to_numpy() == 0))

    len_0_7 = math.floor((2000 / 100) * 70)
    train, test, _ = ProportionalSplit(train_percentage=0.7)(data, split_id=0)
    assert train.s.shape[0] == len_0_7
    assert test.s.shape[0] == 2000 - len_0_7
    assert count_true(train.s.to_numpy() == 0) == round(0.7 * count_true(data.s.to_numpy() == 0))
    assert count_true(train.y.to_numpy() == 0) == round(0.7 * count_true(data.y.to_numpy() == 0))

    len_0_5 = math.floor((2000 / 100) * 50)
    train, test, _ = ProportionalSplit(train_percentage=0.5)(data, split_id=0)
    assert train.s.shape[0] == len_0_5
    assert test.s.shape[0] == 2000 - len_0_5

    len_0_3 = math.floor((2000 / 100) * 30)
    train, test, _ = ProportionalSplit(train_percentage=0.3)(data, split_id=0)
    assert train.s.shape[0] == len_0_3
    assert test.s.shape[0] == 2000 - len_0_3

    len_0_1 = math.floor((2000 / 100) * 10)
    train, test, _ = ProportionalSplit(train_percentage=0.1)(data, split_id=0)
    assert train.s.shape[0] == len_0_1
    assert test.s.shape[0] == 2000 - len_0_1

    len_0_0 = math.floor((2000 / 100) * 0)
    train, test, _ = ProportionalSplit(train_percentage=0.0)(data, split_id=0)
    assert train.s.shape[0] == len_0_0
    assert train.name == "Toy - Train"
    assert test.s.shape[0] == 2000 - len_0_0
    assert test.name == "Toy - Test"


def test_random_seed():
    """test random seed"""
    data: DataTuple = load_data(Toy())
    train_test_0: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train_0, test_0 = train_test_0
    assert train_0 is not None
    assert test_0 is not None
    assert train_0.x.shape[0] > test_0.x.shape[0]
    assert train_0.x['a1'].values[0] == 2.7839204628851526
    assert train_0.x['a2'].values[0] == 6.083679468491366

    assert train_0.x.shape[0] == train_0.s.shape[0]
    assert train_0.s.shape[0] == train_0.y.shape[0]

    train_test_1: Tuple[DataTuple, DataTuple] = train_test_split(data, random_seed=1)
    train_1, test_1 = train_test_1
    assert train_1 is not None
    assert test_1 is not None
    assert train_1.x.shape[0] > test_1.x.shape[0]
    assert train_1.x['a1'].values[0] == 2.6158100167119773
    assert train_1.x['a2'].values[0] == -2.026281606912103

    assert train_1.x.shape[0] == train_1.s.shape[0]
    assert train_1.s.shape[0] == train_1.y.shape[0]

    train_test_2: Tuple[DataTuple, DataTuple] = train_test_split(data, random_seed=2)
    train_2, test_2 = train_test_2
    assert train_2 is not None
    assert test_2 is not None
    assert train_2.x.shape[0] > test_2.x.shape[0]
    assert train_2.x['a1'].values[0] == 0.6431486026206228
    assert train_2.x['a2'].values[0] == -0.09879806963941018

    assert train_2.x.shape[0] == train_2.s.shape[0]
    assert train_2.s.shape[0] == train_2.y.shape[0]

    train_test_3: Tuple[DataTuple, DataTuple] = train_test_split(data, random_seed=3)
    train_3, test_3 = train_test_3
    assert train_3 is not None
    assert test_3 is not None
    assert train_3.x.shape[0] > test_3.x.shape[0]
    assert train_3.x['a1'].values[0] == 0.8165458710908045
    assert train_3.x['a2'].values[0] == -5.456548268244318

    assert train_3.x.shape[0] == train_3.s.shape[0]
    assert train_3.s.shape[0] == train_3.y.shape[0]


def test_binning():
    """test binning"""
    data: DataTuple = load_data(Adult())

    binned: DataTuple = bin_cont_feats(data)

    assert len([col for col in binned.x.columns if col not in data.x.columns]) == 25
    assert 'age' not in binned.x.columns


def test_sequential_split():
    """test sequential split"""
    data: DataTuple = load_data(Toy())
    train: DataTuple
    test: DataTuple
    train, test, _ = SequentialSplit(train_percentage=0.8)(data)
    assert all(data.x.iloc[0] == train.x.iloc[0])
    assert all(data.x.iloc[-1] == test.x.iloc[-1])
    assert len(train) == 1600
    assert len(test) == 400


def test_biased_split():
    """test biased split"""
    data = DataTuple(
        x=pd.DataFrame([0] * 1000, columns=['feat1-']),
        s=pd.DataFrame([1] * 750 + [0] * 250, columns=["sens="]),
        y=pd.DataFrame([1] * 500 + [0] * 250 + [1] * 125 + [0] * 125, columns=["label<"]),
        name="TestData",
    )

    # =================================== mixing factor = 0 =======================================
    biased1, subset = get_biased_subset(data, mixing_factor=0.0, unbiased_pcnt=0.5)
    # expected behavior: in biased1, s=y everywhere; `subset` is just a subset of `data`

    assert biased1.s.shape == (313, 1)
    assert biased1.y.shape == (313, 1)
    assert (biased1.s.to_numpy() == biased1.y.to_numpy()).all()
    assert biased1.name == "TestData - Biased (tm=0.0)"

    # size is exactly 0.5 of the total
    assert subset.s.shape[0] == approx(500, abs=4)  # can be off by 4 because of proportional split
    assert subset.y.shape[0] == approx(500, abs=4)
    assert subset.name == "TestData - Subset (tm=0.0)"

    # the fraction of `data`points where y=s should be the same in the data and the subset
    subset_mean = (subset.s.to_numpy() == subset.y.to_numpy()).mean()
    data_mean = (data.s.to_numpy() == data.y.to_numpy()).mean()
    assert subset_mean == approx(data_mean, abs=0.01)

    biased2, debiased = get_biased_and_debiased_subsets(
        data, mixing_factor=0.0, unbiased_pcnt=0.5, fixed_unbiased=True
    )
    # expected behavior: in biased2, s=y everywhere; in debiased, 50% s=y and 50% s!=y

    assert biased2.s.shape == (313, 1)
    assert biased2.y.shape == (313, 1)
    assert (biased2.s.to_numpy() == biased2.y.to_numpy()).all()

    # assert debiased.s.shape == (626, 1)
    # assert debiased.y.shape == (626, 1)
    assert debiased.s.shape == (374, 1)
    assert debiased.y.shape == (374, 1)
    # for the debiased subset, s=y half of the time
    count = count_true(debiased.s.to_numpy() == debiased.y.to_numpy())
    # assert count == 626 // 2
    assert count == 374 // 2

    # ================================= mixing factor = 0.5 =======================================
    biased1, subset = get_biased_subset(data, mixing_factor=0.5, unbiased_pcnt=0.5)
    # expected behavior: biased1 and `subset` are both just subsets of `data`

    assert biased1.s.shape[0] == approx(500, abs=4)
    assert biased1.y.shape[0] == approx(500, abs=4)
    data_mean = (data.s.to_numpy() == data.y.to_numpy()).mean()
    biased1_mean = (biased1.s.to_numpy() == biased1.y.to_numpy()).mean()
    assert biased1_mean == approx(data_mean, abs=0.01)

    assert subset.s.shape[0] == approx(500, abs=4)
    assert subset.y.shape[0] == approx(500, abs=4)

    # the fraction of `data`points where y=s should be the same in the data and the subset
    subset_mean = (subset.s.to_numpy() == subset.y.to_numpy()).mean()
    assert subset_mean == approx(data_mean, abs=0.01)

    biased2, debiased = get_biased_and_debiased_subsets(
        data, mixing_factor=0.5, unbiased_pcnt=0.5, fixed_unbiased=True
    )
    # expected behavior: biased2 is just a subset of `data`; in debiased, 50% s=y and 50% s!=y

    assert biased2.s.shape[0] == approx(500, abs=4)
    assert biased2.y.shape[0] == approx(500, abs=4)
    data_mean = (data.s.to_numpy() == data.y.to_numpy()).mean()
    biased2_mean = (biased2.s.to_numpy() == biased2.y.to_numpy()).mean()
    assert biased2_mean == approx(data_mean, abs=0.01)

    # assert debiased.s.shape == (564, 1)
    # assert debiased.y.shape == (564, 1)
    assert debiased.s.shape == (374, 1)
    assert debiased.y.shape == (374, 1)
    # for the debiased subset, s=y half of the time
    count = count_true(debiased.s.to_numpy() == debiased.y.to_numpy())
    # assert count == 564 // 2
    assert count == 374 // 2

    # ================================== mixing factor = 1 ========================================
    biased1, subset = get_biased_subset(data, mixing_factor=1.0, unbiased_pcnt=0.5)
    # expected behavior: in biased1, s!=y everywhere; `subset` is just a subset of `data`

    assert biased1.s.shape == (188, 1)
    assert biased1.y.shape == (188, 1)
    assert (biased1.s.to_numpy() != biased1.y.to_numpy()).all()

    assert subset.s.shape[0] == approx(500, abs=4)
    assert subset.y.shape[0] == approx(500, abs=4)

    # the fraction of `data`points where y=s should be the same in the data and the subset
    subset_mean = (subset.s.to_numpy() == subset.y.to_numpy()).mean()
    assert subset_mean == approx(data_mean, abs=0.01)

    biased2, debiased = get_biased_and_debiased_subsets(
        data, mixing_factor=1.0, unbiased_pcnt=0.5, fixed_unbiased=True
    )
    # expected behavior: in biased2, s!=y everywhere; in debiased, 50% s=y and 50% s!=y

    assert biased2.s.shape == (188, 1)
    assert biased2.y.shape == (188, 1)
    assert (biased2.s.to_numpy() != biased2.y.to_numpy()).all()

    # assert debiased.s.shape == (376, 1)
    # assert debiased.y.shape == (376, 1)
    assert debiased.s.shape == (374, 1)
    assert debiased.y.shape == (374, 1)
    # for the debiased subset, s=y half of the time
    count = count_true(debiased.s.to_numpy() == debiased.y.to_numpy())
    # assert count == 376 // 2
    assert count == 374 // 2


def test_biased_split_sizes():
    """test biased split sizes"""
    data = DataTuple(
        x=pd.DataFrame([0] * 1000, columns=['feat1-']),
        s=pd.DataFrame([1] * 750 + [0] * 250, columns=["sens="]),
        y=pd.DataFrame([1] * 500 + [0] * 250 + [1] * 125 + [0] * 125, columns=["label<"]),
        name="TestData",
    )

    biased1, subset = get_biased_subset(data, mixing_factor=0.0, unbiased_pcnt=0.8)
    assert len(biased1) < len(subset)
    biased1, subset = get_biased_subset(data, mixing_factor=0.0, unbiased_pcnt=0.2)
    assert len(biased1) > len(subset)

    biased2, debiased = get_biased_and_debiased_subsets(
        data, mixing_factor=0.0, unbiased_pcnt=0.8, fixed_unbiased=True
    )
    assert len(biased2) < len(debiased)
    biased2, debiased = get_biased_and_debiased_subsets(
        data, mixing_factor=0.0, unbiased_pcnt=0.2, fixed_unbiased=True
    )
    assert len(biased2) > len(debiased)


def test_biased_split_nonbinary():
    """test biased split nonbinary"""
    # generate data that uses -1 and 1 instead of 0 and 1 for s and y
    data = DataTuple(
        x=pd.DataFrame([0] * 1000, columns=['feat1-']),
        s=pd.DataFrame([1] * 750 + [-1] * 250, columns=["sens="]),
        y=pd.DataFrame([1] * 500 + [-1] * 250 + [1] * 125 + [-1] * 125, columns=["label<"]),
        name="TestData",
    )

    biased1, subset = get_biased_subset(data, mixing_factor=0.5, unbiased_pcnt=0.5)
    assert len(biased1) == approx(len(subset), abs=4)



"""EthicML Tests"""
import sys
from typing import List, Dict, Any, NamedTuple, Tuple
from pathlib import Path
import pandas as pd
import pytest
from pytest import approx

from ethicml.algorithms import run_blocking
from ethicml.algorithms.inprocess import (
    # Agarwal,
    Corels,
    GPyT,
    InAlgorithm,
    InAlgorithmAsync,
    InstalledModel,
    Kamiran,
    # Kamishima,
    LR,
    LRCV,
    LRProb,
    Majority,
    MLP,
    SVM,
    ZafarAccuracy,
    ZafarBaseline,
    ZafarEqOdds,
    ZafarEqOpp,
    ZafarFairness,
    Agarwal,
)
from ethicml.evaluators import CrossValidator, CVResults, evaluate_models_async, run_in_parallel
from ethicml.metrics import Accuracy, AbsCV
from ethicml.utility import Heaviside, DataTuple, TrainTestPair, PathTuple, TestPathTuple
from ethicml.data import load_data, Compas, Toy
from ethicml.preprocessing import train_test_split, query_dt
from ethicml.metrics import Metric
from tests.run_algorithm_test import get_train_test, count_true


class InprocessTest(NamedTuple):
    """Define a test for an inprocess model"""

    name: str
    model: InAlgorithm
    num_pos: int
    num_neg: int


INPROCESS_TESTS = [
    InprocessTest(name="SVM", model=SVM(), num_pos=201, num_neg=199),
    InprocessTest(name="Majority", model=Majority(), num_pos=0, num_neg=400),
    InprocessTest(name="MLP", model=MLP(), num_pos=200, num_neg=200),
    InprocessTest(name="Logistic Regression, C=1.0", model=LR(), num_pos=211, num_neg=189),
    InprocessTest(name="LRCV", model=LRCV(), num_pos=211, num_neg=189),
]


@pytest.mark.parametrize("name,model,num_pos,num_neg", INPROCESS_TESTS)
def test_inprocess(name: str, model: InAlgorithm, num_pos: int, num_neg: int):
    """Test an inprocess model"""
    train, test = get_train_test()

    assert isinstance(model, InAlgorithm)
    assert model is not None
    assert model.name == name

    predictions: pd.DataFrame = model.run(train, test)
    assert count_true(predictions.values == 1) == num_pos
    assert count_true(predictions.values == -1) == num_neg


def test_corels(toy_train_test: TrainTestPair) -> None:
    """test corels"""
    model: InAlgorithm = Corels()
    assert model is not None
    assert model.name == "CORELS"

    train_toy, test_toy = toy_train_test
    with pytest.raises(RuntimeError):
        model.run(train_toy, test_toy)

    data: DataTuple = load_data(Compas())
    train, test = train_test_split(data)

    predictions: pd.DataFrame = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 428
    assert predictions.values[predictions.values == 0].shape[0] == 806


def test_cv_svm():
    """test cv svm"""
    train, test = get_train_test()

    hyperparams: Dict[str, List[Any]] = {'C': [1, 10, 100], 'kernel': ['rbf', 'linear']}

    svm_cv = CrossValidator(SVM, hyperparams, folds=3)

    assert svm_cv is not None
    assert isinstance(svm_cv.model(), InAlgorithm)

    cv_results = svm_cv.run(train)

    best_model = cv_results.best(Accuracy())

    predictions: pd.DataFrame = best_model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189


def test_cv_lr(toy_train_test: TrainTestPair) -> None:
    """test cv lr"""
    train, test = toy_train_test

    hyperparams: Dict[str, List[float]] = {'C': [0.01, 0.1, 1.0]}

    lr_cv = CrossValidator(LR, hyperparams, folds=3)

    assert lr_cv is not None
    assert isinstance(lr_cv.model(), InAlgorithm)

    measure = Accuracy()
    cv_results: CVResults = lr_cv.run(train, measures=[measure])

    assert cv_results.best_hyper_params(measure)['C'] == 0.1
    best_model = cv_results.best(measure)

    predictions: pd.DataFrame = best_model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189


def test_parallel_cv_lr(toy_train_test: TrainTestPair) -> None:
    """test parallel cv lr"""
    train, test = toy_train_test

    hyperparams: Dict[str, List[float]] = {'C': [0.001, 0.01]}

    lr_cv = CrossValidator(LR, hyperparams, folds=2, max_parallel=1)

    assert lr_cv is not None
    assert isinstance(lr_cv.model(), InAlgorithm)

    measure = Accuracy()
    cv_results: CVResults = run_blocking(lr_cv.run_async(train, measures=[measure]))

    assert cv_results.best_hyper_params(measure)['C'] == 0.01
    best_model = cv_results.best(measure)

    predictions: pd.DataFrame = best_model.run(train, test)
    assert count_true(predictions.values == 1) == 212
    assert count_true(predictions.values == -1) == 188


def test_fair_cv_lr(toy_train_test: TrainTestPair) -> None:
    """test fair cv lr"""
    train, _ = toy_train_test

    hyperparams: Dict[str, List[float]] = {'C': [1, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6]}

    lr_cv = CrossValidator(LR, hyperparams, folds=3)

    assert lr_cv is not None
    assert isinstance(lr_cv.model(), InAlgorithm)

    primary = Accuracy()
    fair_measure = AbsCV()
    cv_results = lr_cv.run(train, measures=[primary, fair_measure])

    best_result = cv_results.get_best_in_top_k(primary, fair_measure, top_k=3)
    print(best_result)

    assert best_result.params['C'] == 1e-5
    assert best_result.scores['Accuracy'] == approx(0.8475, rel=1e-4)
    assert best_result.scores['CV absolute'] == approx(0.6654, rel=1e-4)


# @pytest.fixture(scope="module")
# def kamishima():
#     kamishima_algo = Kamishima()
#     yield kamishima_algo
#     print("teardown Kamishima")
#     kamishima_algo.remove()


# def test_kamishima(kamishima):
#     train, test = get_train_test()

#     model: InAlgorithmAsync = kamishima
#     assert model.name == "Kamishima"

#     assert model is not None

#     predictions: pd.DataFrame = run_blocking(model.run_async(train, test))
#     assert predictions.values[predictions.values == 1].shape[0] == 208
#     assert predictions.values[predictions.values == -1].shape[0] == 192


@pytest.fixture(scope="module")
def zafar_models():
    """Create and destroy Zafar models

    Returns:
        Tuple of three Zafar model classes
    """
    zafarAcc_algo = ZafarAccuracy
    zafarBas_algo = ZafarBaseline
    zafarFai_algo = ZafarFairness
    yield (zafarAcc_algo, zafarBas_algo, zafarFai_algo)
    print("teardown Zafar Accuracy")
    zafarAcc_algo().remove()


def test_zafar(zafar_models, toy_train_test: TrainTestPair) -> None:
    """

    Args:
        zafarModels:
        toy_train_test:
    """
    train, test = toy_train_test

    model: InAlgorithmAsync = zafar_models[0]()
    assert model.name == "ZafarAccuracy, γ=0.5"

    assert model is not None

    predictions: pd.DataFrame = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 206
    assert predictions.values[predictions.values == -1].shape[0] == 194

    predictions = run_blocking(model.run_async(train, test))
    assert predictions.values[predictions.values == 1].shape[0] == 206
    assert predictions.values[predictions.values == -1].shape[0] == 194

    hyperparams: Dict[str, List[float]] = {'gamma': [1, 1e-1, 1e-2]}

    model = zafar_models[0]
    zafar_cv = CrossValidator(model, hyperparams, folds=3)

    assert zafar_cv is not None
    assert isinstance(zafar_cv.model(), InAlgorithm)

    primary = Accuracy()
    fair_measure = AbsCV()
    cv_results: CVResults = zafar_cv.run(train, measures=[primary, fair_measure])

    best_result = cv_results.get_best_in_top_k(primary, fair_measure, top_k=3)

    assert best_result.params['gamma'] == 1
    assert best_result.scores['Accuracy'] == approx(0.7094, rel=1e-4)
    assert best_result.scores['CV absolute'] == approx(0.9822, rel=1e-4)

    model = zafar_models[1]()
    assert model.name == "ZafarBaseline"

    assert model is not None

    predictions = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189

    predictions = run_blocking(model.run_async(train, test))
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189

    model = zafar_models[2]()
    assert model.name == "ZafarFairness, c=0.001"

    assert model is not None

    predictions = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 215
    assert predictions.values[predictions.values == -1].shape[0] == 185

    predictions = run_blocking(model.run_async(train, test))
    assert predictions.values[predictions.values == 1].shape[0] == 215
    assert predictions.values[predictions.values == -1].shape[0] == 185

    hyperparams = {'c': [1, 1e-1, 1e-2]}

    model = zafar_models[2]
    zafar_cv = CrossValidator(model, hyperparams, folds=3)

    assert zafar_cv is not None
    assert isinstance(zafar_cv.model(), InAlgorithm)

    primary = Accuracy()
    fair_measure = AbsCV()
    cv_results: CVResults = zafar_cv.run(train, measures=[primary, fair_measure])

    best_result = cv_results.get_best_in_top_k(primary, fair_measure, top_k=3)

    assert best_result.params['c'] == 0.01
    assert best_result.scores['Accuracy'] == approx(0.7156, rel=1e-4)
    assert best_result.scores['CV absolute'] == approx(0.9756, rel=1e-4)

    # ==================== Zafar Equality of Opportunity ========================
    zafar_eq_opp: InAlgorithm = ZafarEqOpp()
    assert zafar_eq_opp.name == "ZafarEqOpp, τ=5.0, μ=1.2"

    predictions = zafar_eq_opp.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 217

    # ==================== Zafar Equalised Odds ========================
    zafar_eq_odds: InAlgorithm = ZafarEqOdds()
    assert zafar_eq_odds.name == "ZafarEqOdds, τ=5.0, μ=1.2"

    predictions = zafar_eq_odds.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 189


def test_gpyt_creation():
    """test gpyt creation"""
    gpyt = GPyT(code_dir="non existing path")
    assert gpyt.name == "GPyT_in_True"


# @pytest.fixture(scope="module")
# def gpyt_models():
#     # the gpyt models are created together because they share a git repository
#     gpyt = GPyT(epochs=2, length_scale=2.4)
#     gpyt_dem_par = GPyTDemPar(epochs=2, length_scale=0.05)
#     gpyt_eq_odds = GPyTEqOdds(epochs=2, tpr=1.0, length_scale=0.05)
#     yield (gpyt, gpyt_dem_par, gpyt_eq_odds)
#     gpyt.remove()  # only has to be called from one of the models because they share a directory
#
#
# def test_gpyt(gpyt_models):
#     train, test = get_train_test()
#
#     baseline: InAlgorithmAsync = gpyt_models[0]
#     dem_par: InAlgorithmAsync = gpyt_models[1]
#     eq_odds: InAlgorithmAsync = gpyt_models[2]
#
#     assert baseline.name == "GPyT_in_True"
#     predictions: pd.DataFrame = run_blocking(baseline.run_async(train, test))
#     assert predictions.values[predictions.values == 1].shape[0] == approx(210, rel=0.1)
#     assert predictions.values[predictions.values == -1].shape[0] == approx(190, rel=0.1)
#
#     assert dem_par.name == "GPyT_dem_par_in_True"
#     predictions = run_blocking(dem_par.run_async(train, test))
#     assert predictions.values[predictions.values == 1].shape[0] == approx(182, rel=0.1)
#     assert predictions.values[predictions.values == -1].shape[0] == approx(218, rel=0.1)
#
#     assert eq_odds.name == "GPyT_eq_odds_in_True_tpr_1.0"
#     predictions = run_blocking(eq_odds.run_async(train, test))
#     assert predictions.values[predictions.values == 1].shape[0] == approx(179, rel=0.1)
#     assert predictions.values[predictions.values == -1].shape[0] == approx(221, rel=0.1)


def test_local_installed_lr(toy_train_test: TrainTestPair):
    """test local installed lr"""
    train, test = toy_train_test

    class _LocalInstalledLR(InstalledModel):
        def __init__(self):
            super().__init__(dir_name=".", top_dir="", executable=sys.executable)

        @property
        def name(self) -> str:
            return "local installed LR"

        def _script_command(
            self, train_paths: PathTuple, test_paths: TestPathTuple, pred_path: Path
        ) -> (List[str]):
            script = "./tests/local_installed_lr.py"
            return [
                script,
                str(train_paths.x),
                str(train_paths.y),
                str(test_paths.x),
                str(pred_path),
            ]

    model: InAlgorithm = _LocalInstalledLR()
    assert model is not None
    assert model.name == "local installed LR"

    predictions: pd.DataFrame = model.run(train, test)
    assert count_true(predictions.values == 1) == 211
    assert count_true(predictions.values == -1) == 189


def test_agarwal():
    """test agarwal"""
    train, test = get_train_test()

    agarwal_variants: List[InAlgorithmAsync] = []
    model_names: List[str] = []
    expected_results: List[Tuple[int, int]] = []

    agarwal_variants.append(Agarwal())
    model_names.append("Agarwal, LR, DP")
    expected_results.append((146, 254))

    agarwal_variants.append(Agarwal(fairness="EqOd"))
    model_names.append("Agarwal, LR, EqOd")
    expected_results.append((150, 250))

    agarwal_variants.append(Agarwal(classifier="SVM"))
    model_names.append("Agarwal, SVM, DP")
    expected_results.append((119, 281))

    agarwal_variants.append(Agarwal(classifier="SVM", kernel='linear'))
    model_names.append("Agarwal, SVM, DP")
    expected_results.append((141, 259))

    agarwal_variants.append(Agarwal(classifier="SVM", fairness="EqOd"))
    model_names.append("Agarwal, SVM, EqOd")
    expected_results.append((157, 243))

    agarwal_variants.append(Agarwal(classifier="SVM", fairness="EqOd", kernel="linear"))
    model_names.append("Agarwal, SVM, EqOd")
    expected_results.append((150, 250))

    results = run_blocking(
        run_in_parallel(agarwal_variants, [TrainTestPair(train, test)], max_parallel=1)
    )

    for model, results_for_model, model_name, (pred_true, pred_false) in zip(
        agarwal_variants, results, model_names, expected_results
    ):
        assert model.name == model_name
        assert count_true(results_for_model[0].to_numpy() == 1) == pred_true, model_name
        assert count_true(results_for_model[0].to_numpy() == -1) == pred_false, model_name


def test_threaded_agarwal():
    """test threaded agarwal"""
    models: List[InAlgorithmAsync] = [Agarwal(classifier="SVM", fairness="EqOd")]

    class AssertResult(Metric):
        def score(self, prediction, actual):
            return (
                count_true(prediction.values == 1) == 157
                and count_true(prediction.values == -1) == 243
            )

        @property
        def name(self):
            return "assert_result"

    results = run_blocking(
        evaluate_models_async(datasets=[Toy()], inprocess_models=models, metrics=[AssertResult()])
    )
    assert results.data["assert_result"].iloc[0] == True


def test_lr_prob():
    """test lr prob"""
    train, test = get_train_test()

    model: InAlgorithm = LRProb()
    assert model.name == "Logistic Regression Prob, C=1.0"

    heavi = Heaviside()

    predictions: pd.DataFrame = model.run(train, test)
    predictions = predictions.apply(heavi.apply)
    predictions = predictions.replace(0, -1)
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189


def test_kamiran(toy_train_test: TrainTestPair):
    """test kamiran"""
    train, test = toy_train_test

    kamiran_model: InAlgorithm = Kamiran()
    assert kamiran_model is not None
    assert kamiran_model.name == "Kamiran & Calders LR"

    predictions: pd.DataFrame = kamiran_model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 210
    assert predictions.values[predictions.values == -1].shape[0] == 190

    # remove all samples with s=0 & y=1 from the data
    train_no_s0y1 = query_dt(train, "s != 0 | y != 1")
    predictions = kamiran_model.run(train_no_s0y1, test)



"""Simple class to make class labels binary. Useful if a network uses BCELoss for example."""

import numpy as np
import pandas as pd

from ethicml.utility.data_structures import DataTuple

__all__ = ['LabelBinarizer']


def assert_binary_labels(data_tuple: DataTuple) -> None:
    """Assert that datasets only include binary labels."""
    y_col = data_tuple.y.columns[0]
    assert data_tuple.y[y_col].nunique() == 2
    assert (np.unique(data_tuple.y[y_col].to_numpy()) == np.array([0, 1])).all()


class LabelBinarizer:
    """If a dataset has labels [-1,1], then this will make it so the labels = [0,1]."""

    def __init__(self) -> None:
        """Init LabelBinarizer."""
        self.min_val: int
        self.max_val: int

    def adjust(self, dataset: DataTuple) -> DataTuple:
        """Take a datatuple and make the labels [0,1]."""
        y_col = dataset.y.columns[0]
        assert dataset.y[y_col].nunique() == 2

        # make copy of dataset
        dataset = dataset.replace(y=dataset.y.copy())

        self.min_val = dataset.y.to_numpy().min()
        self.max_val = dataset.y.to_numpy().max()

        y_col = dataset.y.columns[0]

        dataset.y[y_col].replace(self.min_val, 0, inplace=True)
        dataset.y[y_col].replace(self.max_val, 1, inplace=True)

        return DataTuple(x=dataset.x, s=dataset.s, y=dataset.y, name=dataset.name)

    def post_only_labels(self, labels: pd.DataFrame) -> pd.DataFrame:
        """Inverse of adjust but only for a DataFrame instead of a DataTuple."""
        y_col = labels.columns[0]
        assert labels[y_col].nunique() == 2

        # make copy of the labels
        labels_copy = labels.copy()

        y_col = labels_copy.columns[0]
        labels_copy[y_col].replace(0, self.min_val, inplace=True)
        labels_copy[y_col].replace(1, self.max_val, inplace=True)
        return labels_copy

    def post(self, dataset: DataTuple) -> DataTuple:
        """Inverse of adjust."""
        transformed_y = self.post_only_labels(dataset.y)
        return dataset.replace(y=transformed_y)



"""For assessing Balanced Classification Rate (BCR)."""

import pandas as pd

from ethicml.common import implements
from ethicml.metrics.tnr import TNR
from ethicml.metrics.tpr import TPR
from ethicml.utility.data_structures import DataTuple
from .metric import Metric


class BCR(Metric):
    """Balanced Classification Rate."""

    @implements(Metric)
    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        tpr_metric = TPR()
        tpr = tpr_metric.score(prediction, actual)

        tnr_metric = TNR()
        tnr = tnr_metric.score(prediction, actual)

        return (tpr + tnr) / 2

    @property
    def name(self) -> str:
        """Getter for the metric name."""
        return "BCR"



"""Beutel's algorithm."""
from typing import List, Sequence, Union, Dict

from ethicml.utility.data_structures import PathTuple, TestPathTuple, FairnessType
from .pre_algorithm import PreAlgorithmAsync
from .interface import flag_interface


class Beutel(PreAlgorithmAsync):
    """Beutel's adversarially learned fair representations."""

    def __init__(
        self,
        fairness: FairnessType = "DP",
        enc_size: Sequence[int] = (40,),
        adv_size: Sequence[int] = (40,),
        pred_size: Sequence[int] = (40,),
        enc_activation: str = "Sigmoid()",
        adv_activation: str = "Sigmoid()",
        batch_size: int = 64,
        y_loss: str = "BCELoss()",
        s_loss: str = "BCELoss()",
        epochs: int = 50,
        adv_weight: float = 1.0,
        validation_pcnt: float = 0.1,
    ):
        """Init Beutel."""
        # pylint: disable=too-many-arguments
        super().__init__()
        self.flags: Dict[str, Union[str, Sequence[int], int, float]] = {
            "fairness": fairness,
            "enc_size": enc_size,
            "adv_size": adv_size,
            "pred_size": pred_size,
            "enc_activation": enc_activation,
            "adv_activation": adv_activation,
            "batch_size": batch_size,
            "y_loss": y_loss,
            "s_loss": s_loss,
            "epochs": epochs,
            "adv_weight": adv_weight,
            "validation_pcnt": validation_pcnt,
        }

    def _script_command(
        self,
        train_paths: PathTuple,
        test_paths: TestPathTuple,
        new_train_paths: PathTuple,
        new_test_paths: TestPathTuple,
    ) -> List[str]:
        args = flag_interface(train_paths, test_paths, new_train_paths, new_test_paths, self.flags)
        return ["-m", "ethicml.implementations.beutel"] + args

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return "Beutel"



"""This module contains evaluators which apply algorithms over datasets and obtain metrics."""

from .cross_validator import CrossValidator, CVResults
from .evaluate_models import evaluate_models, run_metrics, load_results, evaluate_models_async
from .parallelism import *
from .per_sensitive_attribute import (
    metric_per_sensitive_attribute,
    diff_per_sensitive_attribute,
    ratio_per_sensitive_attribute,
    MetricNotApplicable,
)



"""Accuracy and related metrics."""
from typing import Callable, Optional

import pandas as pd
from sklearn.metrics import accuracy_score, f1_score

from ethicml.common import implements
from ethicml.utility.data_structures import DataTuple
from .metric import Metric

__all__ = ["Accuracy", "F1", "SklearnMetric"]


class SklearnMetric(Metric):
    """Wrapper around an sklearn metric."""

    def __init__(
        self,
        sklearn_metric: Callable[[pd.DataFrame, pd.DataFrame], float],
        name: str,
        pos_class: Optional[int] = None,
    ):
        """Init SKLearnMetric."""
        if pos_class is not None:
            super().__init__(pos_class=pos_class)
        else:
            super().__init__()
        self._metric = sklearn_metric
        self._name = name

    @implements(Metric)
    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        return self._metric(actual.y, prediction)

    @property
    def name(self) -> str:
        """Getter for the metric name."""
        return self._name


class Accuracy(SklearnMetric):
    """Classification accuracy."""

    def __init__(self, pos_class: Optional[int] = None):
        """Init Accuracy."""
        super().__init__(accuracy_score, "Accuracy", pos_class=pos_class)


class F1(SklearnMetric):
    """F1 score: harmonic mean of precision and recall."""

    def __init__(self, pos_class: Optional[int] = None):
        """Init F1."""
        super().__init__(f1_score, "F1", pos_class=pos_class)



"""Common variables / consts that make things run smoother."""

import os
from pathlib import Path
from typing import Any, Callable, Type, TypeVar

ROOT_DIR: str = os.path.abspath(os.path.join(os.path.abspath(__file__), os.pardir))
ROOT_PATH: Path = Path(__file__).parent.resolve()

_F = TypeVar('_F', bound=Callable[..., Any])


class implements:  # pylint: disable=invalid-name
    """Mark a function as implementing an interface."""

    def __init__(self, interface: Type):
        """Instantiate the decorator.

        Args:
            interface: the interface that is implemented
        """
        self.interface = interface

    def __call__(self, func: _F) -> _F:
        """Take a function and return it unchanged."""
        super_method = getattr(self.interface, func.__name__, None)
        assert super_method is not None, f"'{func.__name__}' does not exist in {self.interface}"
        assert super_method.__doc__, f"'{super_method}' has no docstring"
        return func



"""Functions that are common to PyTorch models."""
from typing import Tuple, TYPE_CHECKING

import numpy as np
import torch
from torch import Tensor
from torch.utils.data import Dataset

from ethicml.utility.data_structures import DataTuple, TestTuple

if TYPE_CHECKING:
    import pandas as pd  # only needed for type checking


def _get_info(
    data: TestTuple,
) -> Tuple[
    "np.ndarray[np.float32]",
    "np.ndarray[np.float32]",
    int,
    int,
    int,
    "pd.Index[str]",
    "pd.Index[str]",
]:
    features = data.x.to_numpy(dtype=np.float32)
    sens_labels = data.s.to_numpy(dtype=np.float32)
    num = data.s.shape[0]
    xdim = data.x.shape[1]
    sdim = data.s.shape[1]
    x_names = data.x.columns
    s_names = data.s.columns
    return features, sens_labels, num, xdim, sdim, x_names, s_names


class TestDataset(Dataset):
    """Shared Dataset for pytorch models without labels."""

    def __init__(self, data: TestTuple):
        """Init TestDataset."""
        super().__init__()
        self.x, self.s, self.num, self.xdim, self.sdim, self.x_names, self.s_names = _get_info(data)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        """Implement __getitem__ magic method."""
        return self.x[index, ...], self.s[index, ...]

    def __len__(self) -> int:
        """Implement __len__ magic method."""
        return self.num

    @property
    def names(self) -> Tuple["pd.Index[str]", "pd.Index[str]"]:
        """Get tuple of x names and s names."""
        return self.x_names, self.s_names


class CustomDataset(Dataset):
    """Shared Dataset for pytorch models."""

    def __init__(self, data: DataTuple):
        """Init CutomDataset."""
        super().__init__()
        test = data.remove_y()
        self.x, self.s, self.num, self.xdim, self.sdim, self.x_names, self.s_names = _get_info(test)
        self.y = data.y.to_numpy(dtype=np.float32)
        self.ydim = data.y.shape[1]
        self.y_names = data.y.columns

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Implement __getitem__ magic method."""
        return self.x[index, ...], self.s[index, ...], self.y[index, ...]

    def __len__(self) -> int:
        """Implement __len__ magic method."""
        return self.num

    @property
    def names(self) -> Tuple["pd.Index[str]", "pd.Index[str]", "pd.Index[str]"]:
        """Get tuple of x names, s names and y names."""
        return self.x_names, self.s_names, self.y_names


def quadratic_time_mmd(data_first: Tensor, data_second: Tensor, sigma: float) -> Tensor:
    """Calculate MMD betweer 2 tensors of equal size.

    Args:
        data_first:
        data_second:
        sigma:

    Returns:
        Tensor of MMD in each dim.
    """
    xx_gm = data_first @ data_first.t()
    xy_gm = data_first @ data_second.t()
    yy_gm = data_second @ data_second.t()
    x_sqnorms = torch.diagonal(xx_gm)
    y_sqnorms = torch.diagonal(yy_gm)

    def pad_first(x):
        return torch.unsqueeze(x, 0)

    def pad_second(x):
        return torch.unsqueeze(x, 1)

    gamma = 1 / (2 * sigma ** 2)
    # use the second binomial formula
    kernel_xx = torch.exp(-gamma * (-2 * xx_gm + pad_second(x_sqnorms) + pad_first(x_sqnorms)))
    kernel_xy = torch.exp(-gamma * (-2 * xy_gm + pad_second(x_sqnorms) + pad_first(y_sqnorms)))
    kernel_yy = torch.exp(-gamma * (-2 * yy_gm + pad_second(y_sqnorms) + pad_first(y_sqnorms)))

    xx_num = float(kernel_xx.shape[0])
    yy_num = float(kernel_yy.shape[0])

    mmd2 = (
        kernel_xx.sum() / (xx_num * xx_num)
        + kernel_yy.sum() / (yy_num * yy_num)
        - 2 * kernel_xy.sum() / (xx_num * yy_num)
    )
    return mmd2



"""Wrapper for SKLearn implementation of MLP."""
from typing import Optional, Tuple, Dict

from sklearn.neural_network import MLPClassifier
import pandas as pd

from ethicml.common import implements
from ethicml.utility.data_structures import ActivationType, DataTuple, TestTuple
from .in_algorithm import InAlgorithm


ACTIVATIONS: Dict[str, ActivationType] = {
    "identity": "identity",
    "logistic": "logistic",
    "tanh": "tanh",
    "relu": "relu",
}


class MLP(InAlgorithm):
    """Multi-layer Perceptron."""

    def __init__(
        self,
        hidden_layer_sizes: Optional[Tuple[int]] = None,
        activation: Optional[ActivationType] = None,
    ):
        """Init MLP."""
        super().__init__()
        if hidden_layer_sizes is None:
            self.hidden_layer_sizes = MLPClassifier().hidden_layer_sizes
        else:
            self.hidden_layer_sizes = hidden_layer_sizes
        self.activation: ActivationType = (
            MLPClassifier().activation if activation is None else activation
        )

    @implements(InAlgorithm)
    def run(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        clf = select_mlp(self.hidden_layer_sizes, self.activation)
        clf.fit(train.x, train.y.to_numpy().ravel())
        return pd.DataFrame(clf.predict(test.x), columns=["preds"])

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return "MLP"


def select_mlp(hidden_layer_sizes: Tuple[int], activation: ActivationType) -> MLPClassifier:
    """Create MLP model for the given parameters."""
    assert activation in ACTIVATIONS

    return MLPClassifier(
        hidden_layer_sizes=hidden_layer_sizes, activation=activation, random_state=888
    )



"""This module contains metrics which can be applied to prediction results.

Example:
.. code:: python

   from ethicml.metrics import Accuracy, TPR, ProbPos
   from ethicml.evaluators import run_metrics

   run_metrics(predictions, test_data, metrics=[Accuracy, TPR, ProbPos])
"""
from .accuracy import *
from .anti_spur import AS
from .bcr import BCR
from .confusion_matrix import confusion_matrix, LabelOutOfBounds
from .cv import CV, AbsCV
from .hsic import Hsic
from .metric import Metric
from .normalized_mutual_information import NMI
from .npv import NPV
from .ppv import PPV
from .prob_neg import ProbNeg
from .prob_outcome import ProbOutcome
from .prob_pos import ProbPos
from .theil import Theil
from .tnr import TNR
from .tpr import TPR



"""For assessing ProbNeg."""

import pandas as pd

from ethicml.common import implements
from ethicml.utility.data_structures import DataTuple
from .confusion_matrix import confusion_matrix
from .metric import Metric


class ProbNeg(Metric):
    """Probability of negative prediction."""

    @implements(Metric)
    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        t_neg, _, f_neg, _ = confusion_matrix(prediction, actual, pos_cls=self.positive_class)

        return (t_neg + f_neg) / prediction.size

    @property
    def name(self) -> str:
        """Getter for the metric name."""
        return "prob_neg"



"""For assessing NPV."""

import pandas as pd

from ethicml.common import implements
from ethicml.utility.data_structures import DataTuple
from .confusion_matrix import confusion_matrix
from .metric import Metric


class NPV(Metric):
    """Negative predictive value."""

    @implements(Metric)
    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        t_neg, _, f_neg, _ = confusion_matrix(prediction, actual, self.positive_class)

        return t_neg / (t_neg + f_neg)

    @property
    def name(self) -> str:
        """Getter for the metric name."""
        return "NPV"



"""Class to that can be used to wrap your own dataset in a way EthicML can understand."""
from typing import Dict, List
from .dataset import Dataset


class ConfigurableDataset(Dataset):
    """Dataset that is configurable."""

    def __init__(self) -> None:
        """Init configurable datset object."""
        self.filename = ""
        super().__init__()
        self._dataname: str = "No name assigned."
        self._filename: str = ("No filename assigned. " "Use set_filename(<filename>)")
        self._feature_split: Dict[str, List[str]] = {}
        self.cont_features: List[str] = []
        self.disc_features: List[str] = []

    @property
    def name(self) -> str:
        """Getter for the dataset name."""
        return self._dataname

    @name.setter
    def name(self, value: str) -> None:
        """Setter for the dataset name."""
        self._dataname = value

    @property
    def filename(self) -> str:
        """Getter for the filename."""
        return self._filename

    @filename.setter
    def filename(self, value: str) -> None:
        """Setter for the filename."""
        self._filename = value

    @property
    def feature_split(self) -> Dict[str, List[str]]:
        """Getter for split of features."""
        return self._feature_split

    @feature_split.setter
    def feature_split(self, value: Dict[str, List[str]]) -> None:
        """Setter for split of features."""
        self._feature_split = value



from setuptools import setup, find_packages

# We follow Semantic Versioning (https://semver.org/)
_MAJOR_VERSION = "0"
_MINOR_VERSION = "1"
_PATCH_VERSION = "0"

_VERSION_SUFFIX = "alpha.6"

# Example, '0.4.0-rc1'
version = ".".join([_MAJOR_VERSION, _MINOR_VERSION, _PATCH_VERSION])
if _VERSION_SUFFIX:
    version = f"{version}-{_VERSION_SUFFIX}"

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="EthicML",
    version=version,
    author="Predictive Analytics Lab - University of Sussex",
    author_email="olliethomas86@gmail.com",
    description="A toolkit for understanding and researching algorithmic bias",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/predictive-analytics-lab/EthicML",
    packages=find_packages(),
    package_data={"ethicml.data": ["csvs/*.csv"], "ethicml": ["py.typed"]},
    python_requires=">=3.6",
    install_requires=[
        "dataclasses;python_version<\"3.7\"",  # dataclasses are in the stdlib in python>=3.7
        "fairlearn >= 0.4.0",
        "GitPython >= 2.1.11",
        "matplotlib >= 3.0.2",
        "numpy >= 1.14.2",
        "pandas >= 0.24.0",
        "pipenv >= 2018.11.26",
        "pillow < 7.0",
        "pyarrow >= 0.11",
        "scikit_learn >= 0.20.1",
        "seaborn >= 0.9.0",
        "tqdm >= 4.31.1",
        "typed-argument-parser == 1.4",
        "typing-extensions >= 3.7.2",
    ],
    extras_require={
        'ci': [
            "pytest >= 3.3.2",
            "pytest-cov >= 2.6.0",
            "torch >= 1.1.0, <= 1.1.0.post2",
            "torchvision == 0.3.0",
        ],
        # use `pip install EthicML[dev]` to install development packages
        'dev': ["black", "mypy", "pydocstyle", "pylint", "pytest", "pytest-cov", "pre-commit"],
    },
    classifiers=[  # classifiers can be found here: https://pypi.org/classifiers/
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Typing :: Typed",
    ],
    zip_safe=False,
)



"""Wrapper for calling Kamishima model."""
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Union

import numpy as np
import pandas as pd

from ethicml.common import implements
from ethicml.utility import DataTuple, TestTuple
from .installed_model import InstalledModel
from .in_algorithm import InAlgorithmAsync


class Kamishima(InstalledModel):
    """Model that calls Kamishima's code.

    Based on Algo-Fairness
    https://github.com/algofairness/fairness-comparison/blob/master/fairness/algorithms/kamishima/KamishimaAlgorithm.py
    """

    def __init__(self, eta: float = 1.0):
        """Init Kamishima."""
        super().__init__(
            dir_name="kamishima",
            url="https://github.com/predictive-analytics-lab/kamfadm.git",
            top_dir="kamfadm",
        )
        self.eta = eta

    @staticmethod
    def create_file_in_kamishima_format(data: Union[DataTuple, TestTuple], file_path: str) -> None:
        """Create a text file with the data."""
        if isinstance(data, DataTuple):
            result = (
                pd.concat([data.x, data.s, data.y], axis="columns").to_numpy().astype(np.float64)
            )
        else:
            zeros = pd.DataFrame([0 for _ in range(data.x.shape[0])], columns=['y'])
            result = (
                pd.concat([data.x, data.s, zeros], axis="columns").to_numpy().astype(np.float64)
            )
        np.savetxt(file_path, result)

    @implements(InAlgorithmAsync)
    async def run_async(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            train_path = str(tmp_path / "train.txt")
            test_path = str(tmp_path / "test.txt")
            self.create_file_in_kamishima_format(train, train_path)
            self.create_file_in_kamishima_format(test, test_path)
            min_class_label = train.y[train.y.columns[0]].min()
            model_path = str(tmp_path / "model")
            output_path = str(tmp_path / "output.txt")

            # try:
            await self._call_script(
                [
                    str(self._code_path / "train_pr.py"),
                    "-e",
                    str(self.eta),
                    "-i",
                    train_path,
                    "-o",
                    model_path,
                    "--quiet",
                ]
            )

            await self._call_script(
                [
                    str(self._code_path / "predict_lr.py"),
                    "-i",
                    test_path,
                    "-m",
                    model_path,
                    "-o",
                    output_path,
                    "--quiet",
                ]
            )
            output = np.loadtxt(output_path)
            predictions = output[:, 1].astype(np.float32)
            # except RuntimeError:
            #     predictions = np.ones_like(test.y.to_numpy())

        to_return = pd.DataFrame(predictions, columns=["preds"])

        if to_return["preds"].min() != to_return["preds"].max():
            to_return = to_return.replace(to_return["preds"].min(), min_class_label)
        return to_return

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return "Kamishima"



"""Class to describe features of the Adult dataset."""
from typing_extensions import Literal

from .dataset import Dataset


class Adult(Dataset):
    """UCI Adult dataset."""

    def __init__(
        self,
        split: Literal["Sex", "Race", "Race-Sex", "Custom", "Nationality"] = "Sex",
        discrete_only: bool = False,
    ):
        """Inot Adult dataset."""
        super().__init__()
        self.split = split
        self.discrete_only = discrete_only
        self.features = [
            "age",
            "capital-gain",
            "capital-loss",
            "education-num",
            "education_10th",
            "education_11th",
            "education_12th",
            "education_1st-4th",
            "education_5th-6th",
            "education_7th-8th",
            "education_9th",
            "education_Assoc-acdm",
            "education_Assoc-voc",
            "education_Bachelors",
            "education_Doctorate",
            "education_HS-grad",
            "education_Masters",
            "education_Preschool",
            "education_Prof-school",
            "education_Some-college",
            "hours-per-week",
            "marital-status_Divorced",
            "marital-status_Married-AF-spouse",
            "marital-status_Married-civ-spouse",
            "marital-status_Married-spouse-absent",
            "marital-status_Never-married",
            "marital-status_Separated",
            "marital-status_Widowed",
            "native-country_Cambodia",
            "native-country_Canada",
            "native-country_China",
            "native-country_Columbia",
            "native-country_Cuba",
            "native-country_Dominican-Republic",
            "native-country_Ecuador",
            "native-country_El-Salvador",
            "native-country_England",
            "native-country_France",
            "native-country_Germany",
            "native-country_Greece",
            "native-country_Guatemala",
            "native-country_Haiti",
            "native-country_Holand-Netherlands",
            "native-country_Honduras",
            "native-country_Hong",
            "native-country_Hungary",
            "native-country_India",
            "native-country_Iran",
            "native-country_Ireland",
            "native-country_Italy",
            "native-country_Jamaica",
            "native-country_Japan",
            "native-country_Laos",
            "native-country_Mexico",
            "native-country_Nicaragua",
            "native-country_Outlying-US(Guam-USVI-etc)",
            "native-country_Peru",
            "native-country_Philippines",
            "native-country_Poland",
            "native-country_Portugal",
            "native-country_Puerto-Rico",
            "native-country_Scotland",
            "native-country_South",
            "native-country_Taiwan",
            "native-country_Thailand",
            "native-country_Trinadad&Tobago",
            "native-country_United-States",
            "native-country_Vietnam",
            "native-country_Yugoslavia",
            "occupation_Adm-clerical",
            "occupation_Armed-Forces",
            "occupation_Craft-repair",
            "occupation_Exec-managerial",
            "occupation_Farming-fishing",
            "occupation_Handlers-cleaners",
            "occupation_Machine-op-inspct",
            "occupation_Other-service",
            "occupation_Priv-house-serv",
            "occupation_Prof-specialty",
            "occupation_Protective-serv",
            "occupation_Sales",
            "occupation_Tech-support",
            "occupation_Transport-moving",
            "race_Amer-Indian-Eskimo",
            "race_Asian-Pac-Islander",
            "race_Black",
            "race_Other",
            "race_White",
            "relationship_Husband",
            "relationship_Not-in-family",
            "relationship_Other-relative",
            "relationship_Own-child",
            "relationship_Unmarried",
            "relationship_Wife",
            "salary_<=50K",
            "salary_>50K",
            "sex_Female",
            "sex_Male",
            "workclass_Federal-gov",
            "workclass_Local-gov",
            "workclass_Private",
            "workclass_Self-emp-inc",
            "workclass_Self-emp-not-inc",
            "workclass_State-gov",
            "workclass_Without-pay",
        ]

        self.continuous_features = [
            "age",
            "capital-gain",
            "capital-loss",
            "education-num",
            "hours-per-week",
        ]

        if split == "Sex":
            self.sens_attrs = ["sex_Male"]
            self.s_prefix = ["sex"]
            self.class_labels = ["salary_>50K"]
            self.class_label_prefix = ["salary"]
        elif split == "Race":
            self.sens_attrs = [
                "race_Amer-Indian-Eskimo",
                "race_Asian-Pac-Islander",
                "race_Black",
                "race_Other",
                "race_White",
            ]
            self.s_prefix = ["race"]
            self.class_labels = ["salary_>50K"]
            self.class_label_prefix = ["salary"]
        elif split == "Custom":
            pass
        elif split == "Race-Sex":
            self.sens_attrs = [
                "sex_Male",
                "race_Amer-Indian-Eskimo",
                "race_Asian-Pac-Islander",
                "race_Black",
                "race_Other",
                "race_White",
            ]
            self.s_prefix = ["race", "sex"]
            self.class_labels = ["salary_>50K"]
            self.class_label_prefix = ["salary"]
        elif split == "Nationality":
            self.sens_attrs = [
                "native-country_Cambodia",
                "native-country_Canada",
                "native-country_China",
                "native-country_Columbia",
                "native-country_Cuba",
                "native-country_Dominican-Republic",
                "native-country_Ecuador",
                "native-country_El-Salvador",
                "native-country_England",
                "native-country_France",
                "native-country_Germany",
                "native-country_Greece",
                "native-country_Guatemala",
                "native-country_Haiti",
                "native-country_Holand-Netherlands",
                "native-country_Honduras",
                "native-country_Hong",
                "native-country_Hungary",
                "native-country_India",
                "native-country_Iran",
                "native-country_Ireland",
                "native-country_Italy",
                "native-country_Jamaica",
                "native-country_Japan",
                "native-country_Laos",
                "native-country_Mexico",
                "native-country_Nicaragua",
                "native-country_Outlying-US(Guam-USVI-etc)",
                "native-country_Peru",
                "native-country_Philippines",
                "native-country_Poland",
                "native-country_Portugal",
                "native-country_Puerto-Rico",
                "native-country_Scotland",
                "native-country_South",
                "native-country_Taiwan",
                "native-country_Thailand",
                "native-country_Trinadad&Tobago",
                "native-country_United-States",
                "native-country_Vietnam",
                "native-country_Yugoslavia",
            ]
            self.s_prefix = ["native-country"]
            self.class_labels = ["salary_>50K"]
            self.class_label_prefix = ["salary"]
        else:
            raise NotImplementedError

    @property
    def name(self) -> str:
        """Getter for dataset name."""
        return f"Adult {self.split}"

    @property
    def filename(self) -> str:
        """Getter for filename."""
        return "adult.csv"



"""Split into train and test data."""
from abc import ABC, abstractmethod
import itertools
from typing import Tuple, List, Iterator, Dict

import numpy as np
from numpy.random import RandomState
import pandas as pd
from pandas.testing import assert_index_equal

from ethicml.utility.data_structures import DataTuple
from ethicml.utility.data_helpers import shuffle_df

__all__ = [
    'DataSplitter',
    'ProportionalSplit',
    'SequentialSplit',
    'RandomSplit',
    'train_test_split',
]


class DataSplitter(ABC):
    """Base class for classes that split data."""

    @abstractmethod
    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        """Split the given data.

        Args:
            data: the data to split
            split_id: ID that makes this split unique; this affects for example the random seed

        Returns:
            A tuple consisting of two datatuples that are the result of the split and a dictionary
            that contains information about the split that is useful for logging.
        """


class SequentialSplit(DataSplitter):
    """Take the first N samples for train set and the rest for test set; no shuffle."""

    def __init__(self, train_percentage: float):
        """Init Sequential split."""
        self.train_percentage = train_percentage

    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        """Do Sequential split."""
        del split_id
        train_len = round(self.train_percentage * len(data))

        train = data.apply_to_joined_df(lambda df: df.iloc[:train_len].reset_index(drop=True))
        train = train.replace(name=f"{data.name} - Train")

        test = data.apply_to_joined_df(lambda df: df.iloc[train_len:].reset_index(drop=True))
        test = test.replace(name=f"{data.name} - Test")

        assert len(train) + len(test) == len(data)
        return train, test, {}


def train_test_split(
    data: DataTuple, train_percentage: float = 0.8, random_seed: int = 0
) -> Tuple[DataTuple, DataTuple]:
    """Split a data tuple into two datatuple along the rows of the DataFrames.

    Args:
        data: data tuple to split
        train_percentage: percentage for train split
        random_seed: seed to make splitting reproducible

    Returns:
        train split and test split
    """
    # ======================== concatenate the datatuple to one dataframe =========================
    # save the column names for later
    x_columns: pd.Index = data.x.columns
    s_columns: pd.Index = data.s.columns
    y_columns: pd.Index = data.y.columns

    all_data: pd.DataFrame = pd.concat([data.x, data.s, data.y], axis="columns")

    all_data = shuffle_df(all_data, random_state=1)

    # ============================== split the concatenated dataframe =============================
    # permute
    all_data = shuffle_df(all_data, random_state=random_seed)

    # split
    train_len = int(train_percentage * len(all_data))
    all_data_train = all_data.iloc[:train_len]
    all_data_test = all_data.iloc[train_len:]

    assert isinstance(all_data_train, pd.DataFrame)
    assert isinstance(all_data_test, pd.DataFrame)

    all_data_train = all_data_train.reset_index(drop=True)
    all_data_test = all_data_test.reset_index(drop=True)

    # ================================== assemble train and test ==================================
    train: DataTuple = DataTuple(
        x=all_data_train[x_columns],
        s=all_data_train[s_columns],
        y=all_data_train[y_columns],
        name=f"{data.name} - Train",
    )

    test: DataTuple = DataTuple(
        x=all_data_test[x_columns],
        s=all_data_test[s_columns],
        y=all_data_test[y_columns],
        name=f"{data.name} - Test",
    )

    assert isinstance(train.x, pd.DataFrame)
    assert isinstance(test.x, pd.DataFrame)
    assert_index_equal(train.x.columns, x_columns)
    assert_index_equal(test.x.columns, x_columns)

    assert isinstance(train.s, pd.DataFrame)
    assert isinstance(test.s, pd.DataFrame)
    assert_index_equal(train.s.columns, s_columns)
    assert_index_equal(test.s.columns, s_columns)

    assert isinstance(train.y, pd.DataFrame)
    assert isinstance(test.y, pd.DataFrame)
    assert_index_equal(train.y.columns, y_columns)
    assert_index_equal(test.y.columns, y_columns)

    return train, test


class RandomSplit(DataSplitter):
    """Standard train test split."""

    def __init__(self, train_percentage: float = 0.8, start_seed: int = 0):
        """Init RandomSplit.

        Args:
            train_percentage: how much of the data to use for the train split
            start_seed: random seed for the first split
        """
        super().__init__()
        self.start_seed = start_seed
        self.train_percentage = train_percentage

    def _get_seed(self, split_id: int) -> int:
        return self.start_seed + 2410 * split_id

    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        """Do standard split."""
        random_seed = self._get_seed(split_id)
        split_info: Dict[str, float] = {'seed': random_seed}
        return train_test_split(data, self.train_percentage, random_seed) + (split_info,)


class ProportionalSplit(RandomSplit):
    """Split into train and test while preserving the proportion of s and y."""

    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        """Do proportional split."""
        random_seed = self._get_seed(split_id)

        # local random state that won't affect the global state
        random = RandomState(seed=random_seed)

        s_col = data.s.columns[0]
        y_col = data.y.columns[0]

        s_vals: List[int] = list(map(int, data.s[s_col].unique()))
        y_vals: List[int] = list(map(int, data.y[y_col].unique()))

        train_empty: List[int] = []  # this is necessary because mypy is silly
        test_empty: List[int] = []
        train_idx: np.ndarray[np.int64] = np.array(train_empty, dtype=np.int64)
        test_idx: np.ndarray[np.int64] = np.array(test_empty, dtype=np.int64)

        # iterate over all combinations of s and y
        for s, y in itertools.product(s_vals, y_vals):
            # find all indices for this group
            idx: np.ndarray[np.int64]
            idx = ((data.s[s_col] == s) & (data.y[y_col] == y)).to_numpy().nonzero()[0]

            # shuffle and take subsets
            random.shuffle(idx)
            split_idx: int = round(len(idx) * self.train_percentage)
            # append index subsets to the list of train indices
            train_idx = np.concatenate([train_idx, idx[:split_idx]], axis=0)
            test_idx = np.concatenate([test_idx, idx[split_idx:]], axis=0)

        train: DataTuple = DataTuple(
            x=data.x.iloc[train_idx].reset_index(drop=True),
            s=data.s.iloc[train_idx].reset_index(drop=True),
            y=data.y.iloc[train_idx].reset_index(drop=True),
            name=f"{data.name} - Train",
        )

        test: DataTuple = DataTuple(
            x=data.x.iloc[test_idx].reset_index(drop=True),
            s=data.s.iloc[test_idx].reset_index(drop=True),
            y=data.y.iloc[test_idx].reset_index(drop=True),
            name=f"{data.name} - Test",
        )

        # assert that no data points got lost anywhere
        assert len(data) == len(train) + len(test)

        # assert that we (at least approximately) achieved the specified `train_percentage`
        expected_train_len = round(len(data) * self.train_percentage)
        # the maximum error occurs when all the group splits favor train or all favor test
        num_groups = len(s_vals) * len(y_vals)
        assert expected_train_len - num_groups <= len(train) <= expected_train_len + num_groups

        split_info: Dict[str, float] = {'seed': random_seed}
        return train, test, split_info


def fold_data(data: DataTuple, folds: int) -> Iterator[Tuple[DataTuple, DataTuple]]:
    """So much love to sklearn for making their source code open."""
    indices: np.ndarray[np.int64] = np.arange(data.x.shape[0])

    fold_sizes: np.ndarray[np.int32] = np.full(folds, data.x.shape[0] // folds, dtype=np.int32)
    fold_sizes[: data.x.shape[0] % folds] += 1

    current = 0
    for i, fold_size in enumerate(fold_sizes):
        start, stop = current, current + fold_size
        val_inds: np.ndarray[np.int64] = indices[start:stop]
        train_inds = [i for i in indices if i not in val_inds]  # Pretty sure this is inefficient

        train_x = data.x.iloc[train_inds].reset_index(drop=True)
        train_s = data.s.iloc[train_inds].reset_index(drop=True)
        train_y = data.y.iloc[train_inds].reset_index(drop=True)

        assert train_x.shape == (len(train_inds), data.x.shape[1])
        assert train_s.shape == (len(train_inds), data.s.shape[1])
        assert train_y.shape == (len(train_inds), data.y.shape[1])

        val_x = data.x.iloc[val_inds].reset_index(drop=True)
        val_s = data.s.iloc[val_inds].reset_index(drop=True)
        val_y = data.y.iloc[val_inds].reset_index(drop=True)

        assert val_x.shape == (len(val_inds), data.x.shape[1])
        assert val_s.shape == (len(val_inds), data.s.shape[1])
        assert val_y.shape == (len(val_inds), data.y.shape[1])

        yield DataTuple(
            x=train_x, s=train_s, y=train_y, name=f"{data.name} - train fold {i}"
        ), DataTuple(x=val_x, s=val_s, y=val_y, name=f"{data.name} - test fold {i}")

        current = stop
